from tequila.quantumchemistry.qc_base import QuantumChemistryBase
from tequila.quantumchemistry.chemistry_tools import ParametersQC,NBodyTensor,OrbitalData
import pyscf
from pyscf.gto import Mole
from numpy import ndarray
from typing import Callable
from typing import Union

def from_tequila(molecule:QuantumChemistryBase,**kwargs)->pyscf.gto.Mole: #TODO: Change to something like MoleFromTequila
    geometry = molecule.parameters.get_geometry()
    pyscf_geomstring = ""
    for atom in geometry:
        pyscf_geomstring += "{} {} {} {};".format(atom[0], atom[1][0], atom[1][1], atom[1][2])

    if "point_group" in kwargs:
        point_group = kwargs["point_group"]
    else:
        point_group = None

    mol = pyscf.gto.M(
        atom=pyscf_geomstring,
        basis=molecule.parameters.basis_set,
        charge=molecule.parameters.charge,
        verbose=0,
        spin=0,
    )
    if point_group is not None:
        if point_group.lower() != "c1":
            mol.symmetry = True
            mol.symmetry_subgroup = point_group
        else:
            mol.symmetry = False
    else:
        mol.symmetry = True
    mol.symmetry = False

    mol.build(parse_arg=False)
    mol.ao2mo(mo_coeffs=molecule.integral_manager.orbital_coefficients)
    mol.build()
    return mol
    # solve restricted HF
    mf = pyscf.scf.RHF(mol)#.run()
    mf.verbose = False
    if "verbose" in kwargs:
        mf.verbose = kwargs["verbose"]
    mf.mo_coeff = molecule.integral_manager.orbital_coefficients
    # mf.kernel()
    
    return mf


def MoleculeFromPyscf(molecule:Mole,mo_coeff:Union[ndarray,None]=None,transformation:Union[str,Callable,None]=None,active_orbitals:list=None,frozen_orbitals:list=None,*args,**kwargs)->QuantumChemistryBase:
    
    geo = ''
    for i in  range(len(molecule.atom_coords())):
            c = molecule.atom_coord(i,unit="UA")
            at = ''.join(x for x in molecule.atom_symbol(i) if x.isalpha())
            geo += f'{at} {c[0]} {c[1]} {c[2]}\n'
    if len(molecule.basis):
        basis = molecule.basis
    elif 'basis_set' in kwargs: #when reading molden files, the basis name is lost
        basis = kwargs['basis_set']
        kwargs.pop('basis_set')
    else:
        basis = 'custom'
    parameters = ParametersQC(
            basis_set=basis,
            geometry=geo,
            multiplicity=molecule.multiplicity,
            charge=molecule.charge,
        )
    if "orbital_type" not in kwargs:
        if mo_coeff is None:
            kwargs["orbital_type"] = 'hf'
        else: kwargs["orbital_type"] = 'unknown'
    if 'mo_energy' in kwargs:
        orbital_energies = kwargs['mo_energy']
        kwargs.pop('mo_energy')
    else: orbital_energies = [*range(molecule.nao)]
    if mo_coeff is None:
        mf =pyscf.scf.RHF(molecule)
        mf.kernel()
        mo_coeff = mf.mo_coeff
        orbital_energies = mf.mo_energy #would be weird to provide orbital energies but not mo_coeff
    if molecule.irrep_name is not None:
        irreps = pyscf.symm.label_orb_symm(molecule, molecule.irrep_name, molecule.symm_orb, mo_coeff).tolist()
    else: irreps = ['a1' for _ in range(len(mo_coeff))]
    

    orbitals = [OrbitalData(idx_total=idx, irrep=irr, energy=energy) for idx, (irr, energy) in enumerate(zip(irreps, orbital_energies))]
    for irr in {o.irrep for o in orbitals}:
        for i, o in enumerate([o for o in orbitals if o.irrep == irr]):
            o.idx_irrep = i

    h_ao = molecule.intor("int1e_kin") + molecule.intor("int1e_nuc")
    g_ao = molecule.intor("int2e", aosym="s1")
    S = molecule.intor_symmetric("int1e_ovlp")
    g_ao = NBodyTensor(elems=g_ao, ordering="mulliken")
    kwargs["overlap_integrals"] = S
    kwargs["two_body_integrals"] = g_ao
    kwargs["one_body_integrals"] = h_ao
    kwargs["orbital_coefficients"] = mo_coeff
    
    if "nuclear_repulsion" not in kwargs:
                kwargs["nuclear_repulsion"] = molecule.energy_nuc()

    tqmol = QuantumChemistryBase(parameters=parameters,transformation=transformation,active_orbitals=active_orbitals,frozen_orbitals=frozen_orbitals,point_group=molecule.symmetry_subgroup,*args,**kwargs)
    tqmol.integral_manager.orbital_coefficients = mo_coeff
    return tqmol


import numbers

from ..core import CircuitState, CircuitStyle
from . import PairCorrelatorGate
from ..core import Polygon
from typing import List

class SingleExcitation(PairCorrelatorGate):
    """
        i,j correspond to Spin Orbital index --> ((i,j))
    """
    def __init__(self, i, j, angle, control=None, assume_real=True, encoding="JordanWigner", unit_of_pi: bool = False,n_qubits_is_double:bool=False,*args, **kwargs):
        super().__init__(i=i, j=j, angle=angle, control=control, assume_real=assume_real, encoding=encoding,unit_of_pi=unit_of_pi,*args,**kwargs)
        self.n_qubits_is_double = n_qubits_is_double

    def construct_circuit(self, *args, **kwargs):
        return self._dummy.make_excitation_gate(indices=(self.i, self.j), angle=self.angle,
                                                assume_real=self.assume_real, control=self.control, *args, **kwargs)

    def _render(self, state: CircuitState, style: CircuitStyle) -> str:
        spin = {0: Polygon(-3), 1: Polygon(3)}

        # styling
        shape1 = spin[self.i % 2]
        shape2 = spin[self.j % 2]
        tcol = style.primary.text_color
        gcol = style.primary.gate_color

        # angle
        gcol, tcol = self.render_angle(style, gcol, tcol)
        if 'single' in style.personalized:
            gcol = style.personalized['single'].gate_color
            tcol = style.personalized['single'].text_color
        # TODO: param is not used again in the original??
        param = self.angle
        if isinstance(param, numbers.Number):
            param = "{:1.2f}".format(param)

        # rendering
        result = " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.i // ((not self.n_qubits_is_double)+1),shape=shape1, gcol=gcol,tcol="{" + tcol.name + "}",op="")
        result += " a{qubit} P:fill={gcol}:shape={shape} \\textcolor{tcol}{{{op}}} ".format(qubit=self.j // ((not self.n_qubits_is_double)+1),shape=shape2, gcol=gcol,tcol="{" + tcol.name + "}",op="")
        return result
    
    def used_wires(self) -> List[int]:

        xi = self.i// ((not self.n_qubits_is_double)+1)
        xj = self.j// ((not self.n_qubits_is_double)+1)

        used = list(range(min(xi, xj), max(xi, xj) + 1))
        if self.control is not None:
            used.append(self.control// ((not self.n_qubits_is_double)+1))

        return used


from tencirchem.static.ucc import *
from functools import partial
from collections import defaultdict
from time import time
from typing import  Tuple, Callable, Union
from numbers import Number
from copy import deepcopy
import numpy as np
from scipy.optimize import minimize
import pandas as pd
from pyscf.gto.mole import Mole
from pyscf.scf import RHF
import tensorcircuit as tc

from tencirchem.constants import DISCARD_EPS
from tencirchem.utils.misc import scipy_opt_wrap
from tencirchem.utils.circuit import get_circuit_dataframe
from tencirchem.static.engine_ucc import (
    get_civector,
    get_statevector,
    translate_init_state,
)
from tencirchem.static.ci_utils import get_ci_strings, get_ex_bitstring, get_init_civector
from tencirchem.static.evolve_tensornetwork import get_circuit
from sunrise.expval.tcc_engine.engine_braket import get_energy, get_expval,get_expval_and_grad,get_energy_and_grad
from tequila import Objective,Variable,simulate,assign_variable
from tequila.objective.objective import Variables,FixedVariable

class EXPVAL(UCC):
    def __init__(
        self,
        mol: Union[Mole, RHF],
        init_method="mp2",
        active_space=None,
        aslst=None,  
        mo_coeff=None,
        mode="fermion",
        engine=None,
        run_hf=True,
        run_mp2=True,
        run_ccsd=True,
        run_fci=True, 
    ):
        r"""
        Initialize the class with molecular input.

        Parameters
        ----------
        mol: Mole or RHF
            The molecule as PySCF ``Mole`` object or the PySCF ``RHF`` object
        init_method: str, optional
            How to determine the initial amplitude guess. Accepts ``"mp2"`` (default), ``"ccsd"``, ``"fe"``
            and ``"zeros"``.
        active_space: Tuple[int, int], optional
            Active space approximation. The first integer is the number of electrons and the second integer is
            the number or spatial-orbitals. Defaults to None.
        aslst: List[int], optional
            Pick orbitals for the active space. Defaults to None which means the orbitals are sorted by energy.
            The orbital index is 0-based.

            .. note::
                See `PySCF document <https://pyscf.org/user/mcscf.html#picking-an-active-space>`_
                for choosing the active space orbitals. Here orbital index is 0-based, whereas in PySCF by default it
                is 1-based.
        mo_coeff: np.ndarray, optional
            Molecule coefficients. If provided then RHF is skipped.
            Can be used in combination with the ``init_state`` attribute.
            Defaults to None which means RHF orbitals are used.
        mode: str, optional
            How to deal with particle symmetry, such as whether force electrons to pair as hard-core boson (HCB).
            Possible values are ``"fermion"``, ``"qubit"`` and ``"hcb"``.
            Default to ``"fermion"``.
        engine: str, optional
            The engine to run the calculation. See :ref:`advanced:Engines` for details.
        run_hf: bool, optional
            Whether run HF for molecule orbitals. Defaults to ``True``.
            The argument has no effect if ``mol`` is a ``RHF`` object.
        run_mp2: bool, optional
            Whether run MP2 for initial guess and energy reference. Defaults to ``True``.
        run_ccsd: bool, optional
            Whether run CCSD for initial guess and energy reference. Defaults to ``True``.
        run_fci: bool, optional
            Whether run FCI  for energy reference. Defaults to ``True``.

        """
        super().__init__(mol=mol,
        init_method=init_method,
        active_space=active_space,
        aslst=aslst, #if breaking bcs here, try installing tcc-ng from github, not sure why it may install regular tcc, not ng
        mo_coeff=mo_coeff,
        mode=mode,
        engine=engine,
        run_hf=run_hf,
        run_mp2=run_mp2,
        run_ccsd=run_ccsd,
        run_fci=run_fci)
        # circuit related
        self._init_state_bra = None
        self._init_state_ket = None
        self.ex_ops_bra = None
        self.ex_ops_ket = None
        self._init_guess = None #init_guess for al variables, type {"a":1,...}, or [1,2,] if in the same order as self.total_variables
        # optimization related
        self.scipy_minimize_options = None
        # optimization result
        self.opt_res = None
        # for manually set
        self._params_bra = None
        self._params_ket = None
        self._variables_bra = None
        self._variables_ket = None
        delattr(self,'_params')
        # delattr(self,'init_guess')
        delattr(self,'_param_ids')
        delattr(self,'_init_state')

    def get_opt_function(self, with_time: bool = False) -> Union[Callable, Tuple[Callable, float]]:
        """
        Returns the cost function in SciPy format for optimization.
        The gradient is included.
        Basically a wrapper to :func:`expval_and_grad`.

        Parameters
        ----------
        with_time: bool, optional
            Whether return staging time. Defaults to False.

        Returns
        -------
        opt_function: Callable
            The optimization cost function in SciPy format.
        time: float
            Staging time. Returned when ``with_time`` is set to ``True``.
        """
        if self.is_diagonal():
            expval_and_grad = scipy_opt_wrap(partial(self.energy_and_grad, engine=self.engine))
        else:
            expval_and_grad = scipy_opt_wrap(partial(self.expval_and_grad, engine=self.engine))

        time1 = time()
        if tc.backend.name == "jax":
            logger.info("JIT compiling the circuit")
            _ = expval_and_grad(np.zeros(self.n_variables))
            logger.info("Circuit JIT compiled")
        time2 = time()
        if with_time:
            return expval_and_grad, time2 - time1
        return expval_and_grad

    def is_diagonal(self)->bool:
        '''
        Check is bra==ket, also if bra == None -> bra = ket
        '''
        if all([i is None for i in [self.init_state_bra,self.ex_ops_bra,self.params_bra]]):
            return True
        if not (all([ex in self.ex_ops_bra for ex in self.ex_ops_ket]) and all([ex in self.ex_ops_ket for ex in self.ex_ops_bra])):
            return False
        if not np.allclose(self.init_state_bra,self.init_state_ket):
            return False
        if not (all([pa in self.params_bra for pa in self.params_ket]) and all([pa in self.params_ket for pa in self.params_bra])):
            return False
        return True

    def kernel(self) -> float:
        """
        The kernel to perform the VQE algorithm.
        The L-BFGS-B method in SciPy is used for optimization
        and configuration is possible by setting the ``self.scipy_minimize_options`` attribute.

        Returns
        -------
        e: float
            The optimized energy
        """
        if not self.n_variables:
            if self.is_diagonal():
                return self.energy()
            else: return self.expval()
        energy_and_grad, stating_time = self.get_opt_function(with_time=True)
        
        if self.init_guess is None:
            self.init_guess  = np.zeros(self.n_variables)
        # optimization options
        if self.scipy_minimize_options is None:
            # quite strict
            options = {"ftol": 1e1 * np.finfo(tc.rdtypestr).eps, "gtol": 1e2 * np.finfo(tc.rdtypestr).eps}
        else:
            options = self.scipy_minimize_options
            
        logger.info("Begin optimization")
        time1 = time()
        opt_res = minimize(energy_and_grad, x0=self.init_guess, jac=True, method="L-BFGS-B", options=options)
        time2 = time()
        if not opt_res.success:
            logger.warning("Optimization failed. See `.opt_res` for details.")

        opt_res["staging_time"] = stating_time
        opt_res["opt_time"] = time2 - time1
        opt_res["init_guess"] = self.init_guess
        opt_res["e"] = float(opt_res.fun)
        # prepare for future modification
        p = opt_res.x.copy()
        dvar = {self.total_variables[i]:p[i] for i in range(self.n_variables)}
        opt_res['params_ket'] = [map_variables(pa,dvar) for pa in self.variables_ket]
        
        if self.variables_bra is not None:
            opt_res['params_bra'] = [map_variables(pa,dvar) for pa in self.variables_bra]
        self.opt_res = opt_res
        return opt_res.e
    
    def expval(self, angles : Tensor = None, engine: str = None,hamiltonian=None) -> float:
        """
        Evaluate the total Expectation Value.

        Parameters
        ----------
        angles: Tensor, optional
            The circuit Variables value. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        Expectation Value: float
            Total Expectation Value

        See Also
        --------
        civector: Get the configuration interaction (CI) vector.
        statevector: Evaluate the circuit state vector.
        expval_and_grad: Evaluate the total Expectation Value and parameter gradients.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> round(uccsd.energy([0, 0]), 8)  # HF state
        -1.11670614
        """
        if engine is None:
            engine = self.engine
        if self.is_diagonal():
            return self.energy(angles=angles,engine=engine,hamiltonian=hamiltonian)
        self._sanity_check()
        if angles is None:
            angles = []
        angles  = self._check_params_argument(angles)
        if hamiltonian is None:
            hamiltonian, _, engine = self._get_hamiltonian_and_core(engine)
        e , s = get_expval(angles=angles,hamiltonian=hamiltonian, n_qubits=self.n_qubits, n_elec_s=self.n_elec_s,total_variables=self.total_variables,
                       engine= engine,mode=self.mode,ex_ops=self.ex_ops_ket , ex_ops_bra=self.ex_ops_bra, 
                       params=self.variables_ket ,params_bra=self.variables_bra,
                       init_state=self.init_state_ket , init_state_bra=self.init_state_bra)
        return float(e) + self.e_core*s

    def energy(self, angles : Tensor = None, engine: str = None,hamiltonian=None) -> float:
        """
        Evaluate the total Expectation Value.

        Parameters
        ----------
        angles: Tensor, optional
            The circuit Variables value. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        Expectation Value: float
            Total Expectation Value

        See Also
        --------
        civector: Get the configuration interaction (CI) vector.
        statevector: Evaluate the circuit state vector.
        expval_and_grad: Evaluate the total Expectation Value and parameter gradients.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> round(uccsd.energy([0, 0]), 8)  # HF state
        -1.11670614
        """
        if engine is None:
            engine = self.engine
        if not self.is_diagonal():
            return self.expval(angles=angles,engine=engine,hamiltonian=hamiltonian)
        self._sanity_check()
        if angles is None:
            angles = []
        angles  = self._check_params_argument(angles)
        if hamiltonian is None:
            hamiltonian, _, engine = self._get_hamiltonian_and_core(engine)
        e = get_energy(angles=angles,hamiltonian=hamiltonian, n_qubits=self.n_qubits, n_elec_s=self.n_elec_s,total_variables=self.total_variables,
                       engine= engine,mode=self.mode,ex_ops=self.ex_ops_ket,params=self.variables_ket,init_state=self.init_state_ket)
        return float(e) + self.e_core

    def expval_and_grad(self, angles : Tensor = None, engine: str = None,hamiltonian=None) -> Tuple[float, Tensor]:
        """
        Evaluate the total Expectation Value and parameter gradients.

        Parameters
        ----------
        angles: Tensor, optional
            Variables value. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        Expectation Value: float
            Total energy
        grad: Tensor
            The parameter gradients

        See Also
        --------
        civector: Get the configuration interaction (CI) vector.
        statevector: Evaluate the circuit state vector.
        Expectation Value: Evaluate the Operator Expectation Value.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> e, g = uccsd.energy_and_grad([0, 0])
        >>> round(e, 8)
        -1.11670614
        >>> g  # doctest:+ELLIPSIS
        array([..., ...])
        """
        if engine is None:
            engine = self.engine
        if self.is_diagonal():
            return self.energy_and_grad(angles=angles,engine=engine,hamiltonian=hamiltonian)
        self._sanity_check()
        angles  = self._check_params_argument(angles)
        if hamiltonian is None:
            hamiltonian, _, engine = self._get_hamiltonian_and_core(engine)
        e, g ,s= get_expval_and_grad(hamiltonian=hamiltonian, n_qubits=self.n_qubits, angles= angles,total_variables=self.total_variables,
                                   n_elec_s=self.n_elec_s, engine=engine, mode=self.mode, 
                                   ex_ops=self.ex_ops_ket , ex_ops_bra=self.ex_ops_bra,
                                   params=self.variables_ket,params_bra=self.variables_bra,
                                   init_state=self.init_state_ket, init_state_bra= self.init_state_bra) 
        return float(e + self.e_core*s), tc.backend.numpy(g)

    def energy_and_grad(self, angles: Tensor = None, engine: str = None,hamiltonian=None) -> Tuple[float, Tensor]:
            """
            Evaluate the total energy and parameter gradients.

            Parameters
            ----------
            angles: Tensor, optional
                Variables value. Defaults to None, which uses the optimized parameter
                and :func:`kernel` must be called before.
            engine: str, optional
                The engine to use. Defaults to ``None``, which uses ``self.engine``.

            Returns
            -------
            energy: float
                Total energy
            grad: Tensor
                The parameter gradients

            See Also
            --------
            civector: Get the configuration interaction (CI) vector.
            statevector: Evaluate the circuit state vector.
            energy: Evaluate the total energy.

            Examples
            --------
            >>> from tencirchem import UCCSD
            >>> from tencirchem.molecule import h2
            >>> uccsd = UCCSD(h2)
            >>> e, g = uccsd.energy_and_grad([0, 0])
            >>> round(e, 8)
            -1.11670614
            >>> g  # doctest:+ELLIPSIS
            array([..., ...])
            """
            if engine is None:
                engine = self.engine
            if not self.is_diagonal():
                return self.expval_and_grad(angles=angles,engine=engine,hamiltonian=hamiltonian)
            self._sanity_check()
            angles = self._check_params_argument(angles)
            if hamiltonian is None:
                hamiltonian, _, engine = self._get_hamiltonian_and_core(engine)
            e, g = get_energy_and_grad(
            angles=angles, hamiltonian=hamiltonian,n_qubits=self.n_qubits,n_elec_s= self.n_elec_s, 
            total_variables=self.total_variables,params=self.variables_ket,ex_ops=self.ex_ops_ket, 
            mode=self.mode, init_state=self.init_state_ket, engine=engine   
            )
            return float(e + self.e_core), tc.backend.numpy(g)

    def _check_params_argument(self, params, strict=True):
        if params is None:
            if self.params is not None:
                params = self.params
            else:
                if strict:
                    raise ValueError("Run the `.kernel` method to determine the parameters first")
                else:
                    if self.init_guess  is not None :
                        params = self.init_guess
                    else:
                        params = np.zeros(self.n_variables)
        if len(params) != self.n_variables:
            raise ValueError(f"Incompatible parameter shape. {self.n_variables} is desired. Got {len(params)}")
        return tc.backend.convert_to_tensor(params).astype(tc.rdtypestr)
    
    def _sanity_check(self):
        pass
               
    def civector(self, params: Tensor = None, engine: str = None,ket: bool = True) -> Tensor:
        """
        Evaluate the configuration interaction (CI) vector.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        civector: Tensor
            Corresponding CI vector

        See Also
        --------
        statevector: Evaluate the circuit state vector.
        energy: Evaluate the total energy.
        energy_and_grad: Evaluate the total energy and parameter gradients.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> uccsd.civector([0, 0])  # HF state
        array([1., 0., 0., 0.])
        """
        self._sanity_check()
        params = self._check_params_argument(params)
        self._check_engine(engine)
        d = {self.var_to_param[self.total_variables[i]]:params[i] for i in range(len(params))}
        if engine is None:
            engine = self.engine
        if ket:
            ex_ops  = self.ex_ops_ket
            init_state = self.init_state_ket 
            params = [map_variables(p,d) for p in self.variables_ket]
        else:
            ex_ops  = self.ex_ops_bra
            init_state = self.init_state_bra
            params = [map_variables(p,d) for p in self.variables_bra]
        civector = get_civector(
            params, self.n_qubits, self.n_elec_s, ex_ops,[*range(len(ex_ops))], self.mode, init_state, engine
        )
        return civector

    def _statevector_to_civector(self, statevector=None,ket: bool = True):
        if statevector is None:
            civector = self.civector(ket=ket)
        else:
            if len(statevector) == self.statevector_size:
                ci_strings = self.get_ci_strings()
                civector = statevector[ci_strings]
            else:
                if len(statevector) == self.civector_size:
                    civector = statevector
                else:
                    raise ValueError(f"Incompatible statevector size: {len(statevector)}")

        civector = tc.backend.numpy(tc.backend.convert_to_tensor(civector))
        return civector
    # since there's ci_vector method
    ci_strings = get_ci_strings

    def statevector(self, params: Tensor = None, engine: str = None, ket: bool = True) -> Tensor:
        """
        Evaluate the circuit state vector.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        statevector: Tensor
            Corresponding state vector

        See Also
        --------
        civector: Evaluate the configuration interaction (CI) vector.
        energy: Evaluate the total energy.
        energy_and_grad: Evaluate the total energy and parameter gradients.

        Examples
        --------
        >>> from tencirchem import UCCSD
        >>> from tencirchem.molecule import h2
        >>> uccsd = UCCSD(h2)
        >>> uccsd.statevector([0, 0])  # HF state
        array([0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.])
        """
        self._sanity_check()
        params = self._check_params_argument(params)
        self._check_engine(engine)
        if engine is None:
            engine = self.engine
        if ket:
            ex_ops = self.ex_ops_ket
            init_state = self.init_state_ket
        else:
            ex_ops = self.ex_ops_bra
            init_state = self.init_state_bra
        statevector = get_statevector(
            params, self.n_qubits, self.n_elec_s, ex_ops, [*range(len(ex_ops))], self.mode, init_state, engine
        )
        return statevector

    def get_circuit(self, params: Tensor = None, decompose_multicontrol: bool = False, trotter: bool = False, ket: bool = True) -> tc.Circuit:
        """
        Get the circuit as TensorCircuit ``Circuit`` object

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter.
            If :func:`kernel` is not called before, the initial guess is used.
        decompose_multicontrol: bool, optional
            Whether decompose the Multicontrol gate in the circuit into CNOT gates.
            Defaults to False.
        trotter: bool, optional
            Whether Trotterize the UCC factor into Pauli strings.
            Defaults to False.
        ket: bool, optional
            Whether to return ket or bra circuit
        Returns
        -------
        circuit: :class:`tc.Circuit`
            The quantum circuit.
        """
        if ket:
            ex_ops = self.ex_ops_ket
            init_state = self.init_state_ket 
        else:
            ex_ops = self.ex_ops_bra
            init_state = self.init_state_bra
        if ex_ops is None:
            raise ValueError("Excitation operators not defined")
        params = self._check_params_argument(params, strict=False)
        return get_circuit(params,self.n_qubits,self.n_elec_s,ex_ops,[*range(len(ex_ops))],self.mode,init_state,decompose_multicontrol=decompose_multicontrol,trotter=trotter)

    def print_circuit(self,ket:bool=True):
        """
        Prints the circuit information. If you wish to print the circuit diagram,
        use :func:`get_circuit` and then call ``draw()`` such as ``print(ucc.get_circuit().draw())``.
        """
        c = self.get_circuit(ket=ket)
        df = get_circuit_dataframe(c)

        def format_flop(f):
            return f"{f:.3e}"

        formatters = {"flop": format_flop}
        print(df.to_string(index=False, formatters=formatters))

    def get_init_state_dataframe(self, init_state=None, coeff_epsilon: float = DISCARD_EPS,ket:bool=True) -> pd.DataFrame:
        """
        Returns initial state information dataframe.

        Parameters
        ----------
        coeff_epsilon: float, optional
            The threshold to screen out states with small coefficients.
            Defaults to 1e-12.

        Returns
        -------
        pd.DataFrame

        See Also
        --------
        init_state: The circuit initial state before applying the excitation operators.

        Examples
        --------
        >>> from tencirchem import UCC
        >>> from tencirchem.molecule import h2
        >>> ucc = UCC(h2)
        >>> ucc.init_state = [0.707, 0, 0, 0.707]
        >>> ucc.get_init_state_dataframe()   # doctest: +NORMALIZE_WHITESPACE
             configuration  coefficient
        0          0101        0.707
        1          1010        0.707
        """
        columns = ["configuration", "coefficient"]
        if init_state is None:
            if ket:
                if self.init_state_ket is None:
                    init_state = get_init_civector(self.civector_size)
                else:
                    init_state = self.init_state_ket
            else:
                if self.init_state_bra is None:
                    init_state = get_init_civector(self.civector_size)
                else:
                    init_state = self.init_state_bra
        ci_strings = self.get_ci_strings()
        ci_coeffs = translate_init_state(init_state, self.n_qubits, ci_strings)
        data_list = []
        for ci_string, coeff in zip(ci_strings, ci_coeffs):
            if np.abs(coeff) < coeff_epsilon:
                continue
            ci_string = bin(ci_string)[2:]
            ci_string = "0" * (self.n_qubits - len(ci_string)) + ci_string
            data_list.append((ci_string, coeff))
        return pd.DataFrame(data_list, columns=columns)

    def get_excitation_dataframe(self,ket:bool=True) -> pd.DataFrame:
        '''
        ket:bool Whether to refered to the ket (True) or bra (False)
        '''
        columns = ["excitation", "configuration", "parameter", "initial guess"]
        if ket:
            if self.ex_ops_ket  is None:
                return pd.DataFrame(columns=columns)

            if self.params_ket  is None:
                # optimization not done
                params = [None] * len(self.init_guess)
            else:
                params = self.params_ket 

            ex_ops =self.ex_ops_ket
            init_guess = self.init_guess 
        else:
            if self.ex_ops_bra is None:
                return pd.DataFrame(columns=columns)

            if self.params_bra is None:
                # optimization not done
                params = [None] * len(self.init_guess_bra)
            else:
                params = self.params_bra

            ex_ops =self.ex_ops_bra
            init_guess = self.init_guess_bra
            
        data_list = []

        for i, ex_op in zip([*range(len(ex_ops))], ex_ops):
            bitstring = get_ex_bitstring(self.n_qubits, self.n_elec_s, ex_op, self.mode)
            data_list.append((ex_op, bitstring, params[i], init_guess[i]))
        return pd.DataFrame(data_list, columns=columns)

    def print_excitations(self):
        print('Bra: ',self.get_excitation_dataframe(ket=False).to_string())
        print('Ket: ',self.get_excitation_dataframe(ket=True).to_string())

    def print_ansatz(self):
        df_dict = {
            "#qubits": [self.n_qubits],
            "#params_ket": [self.n_params_ket],
            "#params_bra": [self.n_params_bra],
            "#excitations_ket": [len(self.ex_ops_ket)],
            "#excitations_bra": [len(self.ex_ops_bra)],
        }
        if self.init_state_bra is None:
            df_dict["initial condition bra"] = "RHF"
        else:
            df_dict["initial condition bra"] = "custom"

        if self.init_state is None:
            df_dict["initial condition ket"] = "RHF"
        else:
            df_dict["initial condition ket"] = "custom"
        print(pd.DataFrame(df_dict).to_string(index=False))

    def print_init_state_bra(self):
        print('Bra: ',self.get_init_state_dataframe(self.init_state_bra).to_string())
    
    def print_init_state_ket(self):
        print('Ket: ',self.get_init_state_dataframe(self.init_state_ket).to_string())

    def print_init_state(self):
        self.print_init_state_bra()
        self.print_init_state_ket()
    
    def print_summary(self, include_circuit: bool = False):
        """
        Print a summary of the class.

        Parameters
        ----------
        include_circuit: bool
            Whether include the circuit section.

        """
        print("################################ Ansatz ###############################")
        self.print_ansatz()
        if self.init_state_bra is not None or self.init_state_ket is not None:
            print("############################ Initial Condition ########################")
            if self.init_state_bra is not None:
                self.print_init_state_bra()
            if self.init_state_ket is not None:
                self.print_init_state_ket()
        if include_circuit:
            print("############################### Circuit ###############################")
            self.print_circuit()
        print("############################### Energy ################################")
        self.print_energy()
        print("############################# Excitations #############################")
        self.print_excitations()
        print("######################### Optimization Result #########################")
        if self.opt_res is None:
            print("Optimization not run (.opt_res is None)")
        else:
            print(self.opt_res)

    @property
    def init_state_bra(self) -> Tensor:
        """
        The circuit initial state before applying the excitation operators. Usually RHF.

        See Also
        --------
        get_init_state_dataframe: Returns initial state information dataframe.
        """
        return self._init_state_bra
    
    @property
    def init_state_ket(self) -> Tensor:
        """
        The circuit initial state before applying the excitation operators. Usually RHF.

        See Also
        --------
        get_init_state_dataframe: Returns initial state information dataframe.
        """
        return self._init_state_ket
    
    @property
    def init_state(self) -> Tensor:
        """
        The circuit initial state before applying the excitation operators. Usually RHF.

        See Also
        --------
        get_init_state_dataframe: Returns initial state information dataframe.
        """
        return self._init_state_bra,self._init_state_ket

    @init_state_bra.setter
    def init_state_bra(self, init_state_bra):
        self._init_state_bra = init_state_bra
    
    @init_state_ket.setter
    def init_state_ket(self, init_state_ket):
        self._init_state_ket = init_state_ket
    
    @init_state.setter
    def init_state(self, init_state):
        self._init_state_bra = init_state
        self._init_state_ket = init_state

    @property
    def ex_ops(self) -> Tensor:
        """
        Excitation operators applied to the bra , ket.
        """
        return self.ex_ops_bra , self.ex_ops_ket 

    @ex_ops.setter
    def ex_ops(self, ex_ops):
        self.ex_ops_bra = ex_ops
        self.ex_ops_ket = ex_ops

    @property
    def params_bra(self) -> Tensor:
        """The circuit parameters."""
        if self.opt_res is not None and 'params_bra' in self.opt_res:
            return self.opt_res.params_bra
        elif self._params_bra is not None:
            return self._params_bra
        return None
    
    @property
    def params_ket(self) -> Tensor:
        """The circuit parameters."""
        if self.opt_res is not None and 'params_ket' in self.opt_res:
            return self.opt_res.params_ket
        elif self._params_ket is not None:
            return self._params_ket
        return None
    
    @property
    def params(self) -> Tensor:
        """The circuit parameters."""
        return self.params_bra , self.params_ket
  
    @params_bra.setter
    def params_bra(self, params_bra=None):
        if params_bra is not None:
            params_bra = deepcopy(params_bra)
            var_bra = []
            for i,j in enumerate(params_bra):
                if isinstance(j,Variable):
                    var_bra.append(j)
                    params_bra[i]= j.name
                elif isinstance(j,FixedVariable):
                    var_bra.append(j)
                elif isinstance(j,Objective):
                    var_bra.append(j)
                    params_bra[i] = f'f({j.extract_variables()})'
                else:
                    var_bra.append(assign_variable(j))
            self._params_bra = params_bra
            self._variables_bra = var_bra

    @params_ket.setter
    def params_ket(self, params_ket=None):
        if params_ket is not None:
            params_ket = deepcopy(params_ket)
            var_ket = []
            for i,j in enumerate(params_ket):
                if isinstance(j,Variable):
                    var_ket.append(j)
                    params_ket[i]= j.name
                elif isinstance(j,FixedVariable):
                    var_ket.append(j)
                elif isinstance(j,Objective):
                    var_ket.append(j)
                    params_ket[i] = f'f({j.extract_variables()})'
                else:
                    var_ket.append(assign_variable(j))
            self._params_ket = params_ket
            self._variables_ket = var_ket
    
    @params.setter
    def params(self, params):
        self.params_bra = params
        self.params_ket = params

    @property 
    def param_to_ex_ops_bra(self):
        if self.params_bra is None: return {}
        d = defaultdict(list)
        for i, j in enumerate(self.params_bra):
            d[j].append(self.ex_ops_bra[i])
        return d
    
    @property
    def param_to_ex_ops_ket(self):
        d = defaultdict(list)
        for i, j in enumerate(self.params_ket):
            d[j].append(self.ex_ops_ket[i])
        return d
    
    @property
    def param_to_ex_ops(self):
        db = self.param_to_ex_ops_bra if self.params_bra is not None else {}
        dk = self.param_to_ex_ops_ket
        db.update(dk)
        return db
    
    @property
    def n_variables(self) -> int:
        return len(self.total_variables) if self.total_variables is not None else 0
    
    @property
    def n_variables_bra(self) -> int:
        return len(Objective(self.variables_bra).extract_variables())
                   
    @property
    def n_variables_ket(self) -> int:
        return len(Objective(self.variables_ket).extract_variables())

    @property
    def variables_ket(self) -> List[Variable]:
        return self._variables_ket

    @property
    def variables_bra(self):
        return self._variables_bra
    
    @property
    def variables(self):
        return self._variables_bra,self._variables_ket
    
    @property
    def param_to_var_bra(self):
        d = {}
        for i in range(len(self.params_bra)):
            d[self.params_bra[i]] = self.variables_bra[i]
        return d
    
    @property
    def param_to_var_ket(self):
        d = {}
        for i in range(len(self.params_ket)):
            d[self.params_ket[i]] = self.variables_ket[i]
        return d
    
    @property
    def param_to_var(self):
        d = self.param_to_var_bra
        d.update(self.param_to_var_ket)
        return d

    @property
    def var_to_param_bra(self):
        if self.params_bra is not None:
            d = {}
            for i in range(len(self.params_bra)):
                d[self.variables_bra[i]] = self.params_bra[i] 
            return d
        else: return None
    
    @property
    def var_to_param_ket(self):
        if self.params_ket is not None:
            d = {}
            for i in range(len(self.params_ket)):
                d[self.variables_ket[i]]  =self.params_ket[i]
            return d
    
    @property
    def var_to_param(self):
        if self.var_to_param_bra is None:
            return self.var_to_param_ket
        else: return self.var_to_param_bra.update(self.var_to_param_ket)

    @property
    def total_variables(self) -> Union[list[Variable] , None]   :
        if self.variables_bra is not None and  self.variables_bra is not None:
            return Objective(self.variables_bra+self.variables_ket).extract_variables()
        elif self.variables_bra is None:
            return Objective(self.variables_ket).extract_variables()
        elif self.variables_ket is None:
            return Objective(self.variables_bra).extract_variables()
        else: return None

    @property
    def init_guess(self):
        if isinstance(self._init_guess,dict):
            return [self._init_guess[i] for i in self.total_variables()]
        elif isinstance(self._init_guess,Number):
            return [self._init_guess]
        else: return self._init_guess
        
    @init_guess.setter
    def init_guess(self,init_guess):
        self._init_guess = init_guess

def map_variables(x:Union[Variable,Objective],dvariables:dict):
    if isinstance(x,Variable):
        x = x.map_variables(dvariables)
    elif isinstance(x,Objective):
        x=simulate(x,dvariables)
    return x


from __future__ import annotations
from tequila.circuit._gates_impl import assign_variable
from tequila.circuit.gates import QubitExcitationImpl,X,Phase
from tequila import Variable,BitString
from tequila.quantumchemistry.chemistry_tools import FermionicGateImpl
from tequila.quantumchemistry.qc_base import QuantumChemistryBase
from tequila.utils.exceptions import TequilaException, TequilaWarning
from tequila import assign_variable,QCircuit,QubitWaveFunction
from typing import List,Union,Iterable,Optional,Callable
import copy
from collections import defaultdict
from numpy import ndarray,where,isclose,pi,array
import warnings
import numbers
from copy import deepcopy
import sunrise
from tequila.circuit.gates import X

class FCircuit:
    """
    Fundamental class representing an abstract circuit.

    Attributes
    ----------
    depth:
        returns the gate depth of the circuit.
    gates:
        returns the gates in the circuit, as a list.
    moments:
        returns the circuit as a list of Moment objects.
    n_qubits:
        the number of qubits on which the circuit operates.
    numbering:
        returns the numbering convention use by tequila circuits.
    qubits:
        returns a list of qubits acted upon by the circuit.
    initial_state:
        returns QubitWaveFunction which defines initial state. It must be on UPTHENDOWN
        

    Methods
    -------
    make_parameter_map:


    """

    def __init__(self, gates:list=None, initial_state:Union[QCircuit,QubitWaveFunction,str,int]=None, parameter_map=None):
        """
        init
        Parameters
        ----------
        gates:
            (Default value = None)
            the gates to include in the circuit.
        parameter_map:
            (Default value = None)
            mapping to indicate where in the circuit certain parameters appear.
        """
        self._n_qubits = None
        self._min_n_qubits = 0
        self._initial_state = None
        if gates is None:
            self._gates = []
        else:
            self._gates = list(gates)

        if parameter_map is None:
            self._parameter_map = self.make_parameter_map()
        else:
            self._parameter_map = parameter_map
        self.initial_state = initial_state 
        self.verify()

    def export_to(self, filename: str,*args, **kwargs):
        """
        Export to png, pdf, qpic, tex with qpic backend
        Parameters
        """
        U = deepcopy(self)
        if 'n_orb' in kwargs:
            n_orb = kwargs['n_orb']
        else: n_orb = self.n_qubits//2
        U.to_udud(n_orb)
        if 'n_qubits_is_double' in kwargs:
            n_qubits_is_double = kwargs['n_qubits_is_double']
            kwargs.pop('n_qubits_is_double')
        else: n_qubits_is_double = False
        gU = sunrise.graphical.GraphicalCircuit.from_circuit(U=U, n_qubits_is_double=n_qubits_is_double)
        gU.export_to(filename=filename,*args,**kwargs)
        

    @property
    def initial_state(self)->QubitWaveFunction:
        return self._initial_state

    @initial_state.setter
    def initial_state(self,initial_state=None):
        '''
        IT MUST BE UPTHENDOWN
        '''
        if initial_state is None or isinstance(initial_state,QubitWaveFunction):
            pass
        elif isinstance(initial_state,Union[QCircuit,FCircuit]):
            initial_state = sunrise.simulate(initial_state,variables={})
        elif isinstance(initial_state,str):
            initial_state = QubitWaveFunction.from_string(initial_state)
        elif isinstance(initial_state,(list,ndarray)):
            initial_state = QubitWaveFunction.from_array(initial_state)
        else:
            try:
                initial_state = QubitWaveFunction.convert_from(val=initial_state,n_qubits=self.n_qubits)
            except:
                raise TequilaException(f'Init_state format not recognized, provided {type(initial_state).__name__}')
        if initial_state is not None:
            self.n_qubits = max(self.n_qubits,initial_state.n_qubits)
            initial_state._n_qubits = self.n_qubits
            self._initial_state = initial_state
            self._verify_state()
    @property
    def depth(self):
        """
        gate depth of the abstract circuit.
        Returns
        -------
        int: the depth.

        """
        return len(self.gates)

    @property
    def gates(self)->List[sunrise.fermionic_excitation.fgateimpl.FGateImpl]:
        if self._gates is None:
            return []
        else:
            return self._gates

    @gates.setter
    def gates(self, other):
        self._gates = list(other)

    @property
    def qubits(self):
        accumulate = []
        for g in self.gates:
            accumulate += list(g.qubits)
        if self._initial_state is not None:
            accumulate.extend([*range(self._initial_state.n_qubits)])
        return sorted(list(set(accumulate)))

    @property
    def n_qubits(self):
        idx = max(self.max_qubit() + 1, self._min_n_qubits)
        wvf = self._initial_state.n_qubits if self._initial_state is not None else 0
        return max(idx,wvf)

    @n_qubits.setter
    def n_qubits(self, other):
        self._min_n_qubits = other
        if other < self.max_qubit() + 1:
            raise TequilaException(
                "You are trying to set n_qubits to "
                + str(other)
                + " but your circuit needs at least: "
                + str(self.max_qubit() + 1)
            )
        return self

    @property
    def variables(self)->list:
        v = []
        for gate in self.gates:
            if gate._name == 'UR':
                v.append(gate.variables)
            v.append(gate.variables)
        return v

    def make_parameter_map(self) -> dict:
        """
        Returns
        -------
            ParameterMap of the circuit: A dictionary with
            keys: variables in the circuit
            values: list of all gates and their positions in the circuit
            e.g. result[Variable("a")] = [(3, Rx), (5, Ry), ...]
        """
        parameter_map = defaultdict(list)
        for idx, gate in enumerate(self.gates):
            if gate.is_parameterized():
                variables = gate.extract_variables()
                for variable in variables:
                    parameter_map[variable] += [(idx, gate)]
        return parameter_map

    def is_primitive(self):
        """
        Check if this is a single gate wrapped in this structure
        :return: True if the circuit is just a single gate
        """
        return len(self.gates) == 1

    def replace_gates(self, positions: list, circuits: list, replace: list = None):
        """
        Notes
        ----------
        Replace or insert gates at specific positions into the circuit
        at different positions (faster than multiple calls to replace_gate)

        Parameters
        ----------
        positions: list of int:
            the positions at which the gates should be added. Always refer to the positions in the original circuit
        circuits: list or QCircuit:
            the gates to add at the corresponding positions
        replace: list of bool: (Default value: None)
            Default is None which corresponds to all true
            decide if gates shall be replaced or if the new parts shall be inserted without replacement
            if replace[i] = true: gate at position [i] will be replaced by gates[i]
            if replace[i] = false: gates[i] circuit will be inserted at position [i] (beaming before gate previously at position [i])
        Returns
        -------
            new circuit with inserted gates
        """

        assert len(circuits) == len(positions)
        if replace is None:
            replace = [True] * len(circuits)
        else:
            assert len(circuits) == len(replace)

        dataset = zip(positions, circuits, replace)
        dataset = sorted(dataset, key=lambda x: x[0])

        new_gatelist = []
        last_idx = -1

        for idx, circuit, do_replace in dataset:
            # failsafe
            if hasattr(circuit, "gates"):
                gatelist = circuit.gates
            elif isinstance(circuit, Iterable):
                gatelist = circuit
            else:
                gatelist = [circuit]

            new_gatelist += self.gates[last_idx + 1 : idx]
            new_gatelist += gatelist
            if not do_replace:
                new_gatelist.append(self.gates[idx])

            last_idx = idx

        new_gatelist += self.gates[last_idx + 1 :]

        result = QCircuit(gates=new_gatelist)
        result.n_qubits = max(result.n_qubits, self.n_qubits)
        return result

    def insert_gates(self, positions, gates):
        """
        See replace_gates
        """
        return self.replace_gates(positions=positions, circuits=gates, replace=[False] * len(gates))

    def dagger(self):
        """
        Returns
        ------
        QCircuit:
            The adjoint of the circuit
        """
        result = FCircuit()
        for g in reversed(self.gates):
            result += self.wrap_gate(g.dagger())
        return result

    def extract_variables(self) -> list:
        """
        return a list containing all the variable objects contained in any of the gates within the unitary
        including those nested within gates themselves.

        Returns
        -------
        list:
            the variables of the circuit
        """
        return list(self._parameter_map.keys())

    def is_fully_parametrized(self):
        """
        Returns
        -------
        bool:
            whether or not all gates in the circuit are paremtrized
        """
        for gate in self.gates:
            if not gate.is_parameterized():
                return False
            else:
                if hasattr(gate, "parameter"):
                    if not hasattr(gate.parameter, "wrap"):
                        return False
                    else:
                        continue
                else:
                    continue
        return True

    def is_fully_unparametrized(self):
        """
        Returns
        -------
        bool:
            whether or not all gates in the circuit are unparametrized
        """
        for gate in self.gates:
            if not gate.is_parameterized():
                continue
            else:
                if hasattr(gate, "parameter"):
                    if not hasattr(gate.parameter, "wrap"):
                        continue
                    else:
                        return False
                else:
                    return False
        return True

    def is_mixed(self):
        return not (self.is_fully_parametrized() or self.is_fully_unparametrized())

    def extract_indices(self) -> list:
        l = []
        for gate in self.gates:
            l.extend(gate.indices)
        return l

    def max_qubit(self):
        """
        Returns:
        int:
             Highest index of qubits in the circuit
        """
        qmax = 0
        for g in self.gates:
            qmax = max(qmax, g.max_qubit)
        if self.initial_state is not None:
            qmax = max(qmax,self.initial_state.n_qubits-1)
        return qmax

    def __iadd__(self, other):
        other = self.wrap_gate(gate=other.gates)

        offset = len(self.gates)
        for k, v in other._parameter_map.items():
            self._parameter_map[k] += [(x[0] + offset, x[1]) for x in v]

        self._gates += other.gates
        self._min_n_qubits = max(self._min_n_qubits, other._min_n_qubits)
        if self._initial_state is None:
            self._initial_state = other._initial_state
        elif other._initial_state is not None and other._initial_state != self._initial_state:
            raise TequilaException(f"FermionicCircuit + FermionicCircuit with two different initial states:\n{self._initial_state}, {other._initial_state}")
        return self

    def __add__(self, other):
        other = self.wrap_gate(other.gates)
        gates = [deepcopy(g) for g in (self.gates + other.gates)]
        result = FCircuit(gates=gates)
        result._min_n_qubits = max(self._min_n_qubits, other._min_n_qubits)
        initial_state = self.initial_state
        if self._initial_state is None:
                initial_state = other._initial_state
        elif other._initial_state is not None and other._initial_state != self._initial_state:
            raise TequilaException(f"FermionicCircuit + FermionicCircuit with two different initial states:\n{self._initial_state}, {other._initial_state}")
        result.initial_state = initial_state
        return result

    def __str__(self):
        result = "Fermionic circuit: \n"
        if self.initial_state is not None:
            result += 'with Initial State: \n'
            result += str(self.initial_state) +'\n'
        if len(self.gates):
            result += 'with gates:\n'
            for g in self.gates:
                result += str(g) + "\n"
        return result

    def __eq__(self, other) -> bool:
        if len(self.gates) != len(other.gates):
            return False
        for i,gate in enumerate(self.gates):
            if gate != other.gates[i]:
                return False
        if len(self.initial_state.to_array()) != len(other.initial_state.to_array()):
            return False
        if self.initial_state.numbering != other.initial_state.numbering:
            return False
        return isclose(self.initial_state.inner(other=other.initial_state),1.,atol=1.e-4)

    def __repr__(self):
        return self.__str__()

    def to_upthendown(self,norb)->FCircuit:
        '''
        Initial State can't be reordered, it must always be on upthendown
        '''
        u = []
        for gate in self.gates:
            g = deepcopy(gate).to_upthendown(norb)
            g.reordered = True
            u.append(g)
        return FCircuit(gates=u,parameter_map=self._parameter_map,initial_state=self.initial_state)

    def to_udud(self,norb)->FCircuit:
        '''
        Initial State can't be reordered, it must always be on upthendown
        '''
        u = []
        for gate in self.gates:
            g = deepcopy(gate).to_udud(norb)
            g.reordered = False
            u.append(g)
        return FCircuit(gates=u,parameter_map=self._parameter_map,initial_state=self.initial_state)

    @classmethod
    def from_Qcircuit(cls,circuit:QCircuit,**kwargs):
        operations = FCircuit()
        reference = QCircuit()
        if 'reordered' in kwargs:
            reordered = kwargs['reordered']
        else: reordered = True 
        begining = True
        if 'variables' in kwargs:
            circuit = circuit.map_variables(kwargs['variables'])
            kwargs.pop("variables")
        for gate in circuit.gates:
            if begining and not hasattr(gate,'_parameter') or isinstance(gate._parameter,numbers.Number):
                reference += gate
            elif isinstance(gate,QubitExcitationImpl): #maybe we can consider other gates but basic implementation for the moment
                if isinstance(gate._parameter,numbers.Number):
                    reference += gate
                elif isinstance(gate,FermionicGateImpl):
                    begining = False 
                    operations += sunrise.gates.FermionicExcitation(indices=gate.indices,variables=gate.parameter,reordered=reordered)
                else:
                    temp = []
                    for i in range(len(gate._target)//2):
                        temp.append((gate._target[2*i],gate._target[2*i+1]))
                    operations += sunrise.gates.FermionicExcitation(indices=temp,variables=gate.parameter,reordered=reordered)
            else:
                raise TequilaException(f'Gate {gate._name}({gate._parameter}) not allowed')
        if not reordered and len(reference.gates):
            raise TequilaException('reordered=False only allowed if any initial_state provided due to backend restrictions')
        return cls(gates=operations._gates,initial_state=reference)

    @classmethod
    def from_edges(cls,edges:Union[list,tuple],label=None,n_orb:int=0, use_units_of_pi=False,ladder=True):
        operations = FCircuit()
        if n_orb is not None:
            include_reference = True
            reference = QCircuit()
        else: 
            include_reference = False
            reference = None
        for edge in edges:
            if include_reference:
                reference += X([edge[0],edge[0]+n_orb])
            previous = edge[0]
            for i in edge[1:]:
                v = Variable(((previous,i),'D',label))
                if use_units_of_pi:
                    angle = angle * pi
                operations += sunrise.gates.UC(i=previous,j=i,variables=v)
                if ladder:
                    previous = i
        return cls(gates=operations._gates,initial_state=reference)

    @staticmethod
    def wrap_gate(gate):
        """
        take a gate and return a qcircuit containing only that gate.
        Parameters
        ----------
        gate: QGateImpl
            the gate to wrap in a circuit.

        Returns
        -------
        QCircuit:
            a one gate circuit.
        """
        if isinstance(gate, FCircuit):
            return gate
        if isinstance(gate, list):
            return FCircuit(gates=gate)
        else:
            return FCircuit(gates=[gate])

    def verify(self) -> bool:
        """
        make sure self is built properly and of the correct types.
        Returns
        -------
        bool:
            whether or not the circuit is properly constructed.

        """
        if not len(self._gates) is None and self.initial_state is None:
            return True
        test = []
        for (k,v,) in self._parameter_map.items():
            test = [self.gates[x[0]] == x[1] for x in v]
            test += [k in self._gates[x[0]].extract_variables() for x in v]
        return all(test)

    def map_qubits(self, qubit_map):
        """

        E.G.  Rx(1)Ry(2) --> Rx(3)Ry(1) with qubit_map = {1:3, 2:1}

        Parameters
        ----------
        qubit_map
            a dictionary which maps old to new qubits

        Returns
        -------
        A new circuit with mapped qubits
        """

        new_gates = [gate.map_qubits(qubit_map) for gate in self.gates]
        # could speed up by applying qubit_map to parameter_map here
        # currently its recreated in the init function
        return FCircuit(gates=new_gates)

    def map_variables(self, variables: dict, *args, **kwargs):
        """

        Parameters
        ----------
        variables
            dictionary with old variable names as keys and new variable names or values as values
        Returns
        -------
        Circuit with changed variables

        """

        variables = {assign_variable(k): assign_variable(v) for k, v in variables.items()}

        # failsafe
        my_variables = self.extract_variables()
        for k, v in variables.items():
            if k not in my_variables:
                warnings.warn(
                    "map_variables: variable {} is not part of circuit with variables {}".format(k, my_variables),
                    TequilaWarning,
                )

        new_gates = [copy.deepcopy(gate).map_variables(variables) for gate in self.gates]

        return FCircuit(gates=new_gates,initial_state=self.initial_state)

    def to_matrix(self, variables=None): #TODO:  Should we?
        pass 

    def add_controls(self, control, inpl: Optional[bool] = False) -> Optional[FCircuit]: #TODO
        pass 
      
    def to_qcircuit(self,molecule:QuantumChemistryBase=None,transformation:Callable=None) -> QCircuit: 
        '''
        Transform from FCircuit to Qcircuit
        :param molecule Any kind of Tequila/Sunrise Molecule with molecule.make_excitation_gate function
        :param transformation Callable function such as transformation(FCircuit) -> QCircuit 
        '''
        U = deepcopy(self)
        if molecule is not None:
            assert molecule.transformation is not None
            res = QCircuit()
            U = U.to_udud(molecule.n_orbitals)
            molecule.transformation.upthendown = True
            if self.initial_state is not None:
                idx = where(array(self._initial_state.to_array())>1.e-6)[0]
                if not len(idx):
                    pass
                elif len(idx)>1:
                    warnings.warn("Don't now how to prepare the circuit initial state, skyped for safety",TequilaWarning)
                else:
                    res += X(target=[i for i in range(len(bin(idx[0])[2:])) if bin(idx[0])[2:][i]=='1'])
            for gate in U.gates:
                if gate.name == 'UR':
                    res += molecule.make_excitation_gate(indices=gate.indices[0],angle=gate.variables)#TODO: Control
                    res += molecule.make_excitation_gate(indices=gate.indices[1],angle=gate.variables)#TODO: Control
                elif gate.name in ['UC', 'FermionicExcitation', 'GenericFermionic']:
                    res += molecule.make_excitation_gate(indices=gate.indices[0],angle=gate.variables)#TODO: Control
                elif gate.name == 'Ph':
                    res +=  Phase(target=(gate.indices[0][0][0]//2 + (gate.indices[0][0][0]%2)*molecule.n_orbitals),angle=gate.variables)  #TODO: Control
                else:
                    raise TequilaException(f'Gate {gate} not idenified')
            return res
        elif transformation is not None:
            return transformation(U)
        else:
            raise TequilaException("No Possible to transform from FCircuit to QCircuit without further information")
    
    def _verify_state(self):
        state = self._initial_state.to_array()
        indices = where(array(state)>1.e-6)[0]
        nozero = [bin(i)[2:] for i in indices]
        ne = nozero[0].count('1')
        if not all([st.count('1')==ne for st in nozero]):
            raise TequilaException('The Initial State is not Particle-conserving')
        if not all([isclose(state[i].imag,0.,atol=1.e-6) for i in indices]):
            raise TequilaException('No Imaginary states for Chemistry States')
    

if __name__ == '__main__':
    import tequila as tq
    import sunrise as sun
    
    U = sun.gates.FermionicExcitation([(0,2),(1,3)],"a")
    U.initial_state = tq.gates.X([0,4])
    U += sun.gates.UC(0,1,"b")
    U = U + sun.gates.UR(1,2,'c')
    U += sun.gates.FermionicExcitation([(1,2)],'d',reordered=True)
    print(U)
    # print(U.make_parameter_map())
    # print(U.max_qubit())
    A = U.to_upthendown(4)
    print(A)
    # print(A.extract_indices())
    print(A.variables)
    # print(A.initial_state)
    print(A.extract_variables())
    # print(A.make_parameter_map())
    E = A.to_udud(4)
    print(E)
    B = FCircuit.from_edges([(0,1),(2,3)])
    print(B.extract_variables())
    print(B.extract_indices())
    print(B)
    C = FCircuit.from_edges([(0,1),(2,3)],n_orb=4)
    print(C)



from .pair_correlator import PairCorrelatorGate
from .orbital_rotator import OrbitalRotatorGate
from .single_excitation import SingleExcitation
from .double_excitation import DoubleExcitation
from .generic_exitation import GenericExcitation



from tequila.apps.adapt.adapt import MolecularPool as MainMP


class MolecularPool(MainMP):
    def make_indices_singles(self, generalized=False):
        indices = []
        for p in range(self.molecule.n_electrons // 2):
            for q in range(self.molecule.n_electrons // 2, self.molecule.n_orbitals):
                if(hasattr(self.molecule,"select")):
                    if (self.molecule.select[q] == "F" and self.molecule.select == "F"):
                        indices.append([(2 * p, 2 * q)])
                        indices.append([(2 * p + 1, 2 * q + 1)])
                else:
                    indices.append([(2*p, 2*q)])
                    indices.append([(2*p+1, 2*q+1)])
        if not generalized:
            return indices

        for p in range(self.molecule.n_orbitals):
            for q in range(p + 1, self.molecule.n_orbitals):
                if(hasattr(self.molecule,"select")):
                    if (self.molecule.select[q] == "F" and self.molecule.select == "F"):
                        if [(2 * p, 2 * q)] in indices:
                            continue
                        indices.append([(2 * p, 2 * q)])
                        indices.append([(2 * p + 1, 2 * q + 1)])
                else:
                    if [(2*p, 2*q)] in indices:
                        continue
                    indices.append([(2*p, 2*q)])
                    indices.append([(2*p+1, 2*q+1)])
        return self.sort_and_filter_unique_indices(indices)
    def make_indices_doubles(self, generalized=False, paired=True):
        indices = []
        for p in range(self.molecule.n_electrons//2):
            for q in range(self.molecule.n_electrons//2, self.molecule.n_orbitals):
                indices.append([(2*p, 2*q),(2*p+1, 2*q+1)])

        if not generalized:
            return indices

        for p in range(self.molecule.n_orbitals):
            for q in range(p+1, self.molecule.n_orbitals):
                idx = [(2 * p, 2 * q), (2 * p + 1, 2 * q + 1)]
                if idx in indices:
                    continue
                indices.append(idx)
        if not paired:
            indices += self.make_indices_doubles_all(generalized=generalized)

        return self.sort_and_filter_unique_indices(indices)
    def make_indices_doubles_all(self, generalized=False):
        from itertools import combinations
        singles = self.make_indices_singles(generalized=generalized)
        unwrapped = [x[0] for x in singles]
        # now make all combinations of singles
        if(hasattr(self.molecule,"select")):
            indices = [x for x in combinations(unwrapped, 2) if self.molecule.verify_excitation(x)]
        else:
            indices = [x for x in combinations(unwrapped, 2)]
        return self.sort_and_filter_unique_indices(indices)



import tequila as tq
from tequila import QCircuit,QubitWaveFunction,QubitHamiltonian
from tequila.objective.objective import Variables
from tequila.quantumchemistry.qc_base import QuantumChemistryBase as tqMolecule
import numpy as np
from scipy.linalg import logm, eigh
import itertools
import math
from typing import List,Tuple,Union

class Input_State:
    """
    Wrapper class to hold the input state to measure the operators
    The input state can be a quantum circuit or a wavefunction
    """
    circuit = None
    variables = None
    wavefunction = None

    def __init__(self, circuit=None, variables=None, wavefunction=None):
        if circuit is None and wavefunction is None:
            raise ValueError("Either a circuit or a wavefunction must be provided")
        if circuit is not None and wavefunction is not None and wavefunction != 0:
            raise ValueError("Only one of circuit or wavefunction must be provided")
        
        self.circuit = circuit
        self.variables = variables
        self.wavefunction = wavefunction

    def get_circuit(self):
        """
        Get the circuit from the wavefunction or the circuit
        """
        if self.wavefunction is not None and self.circuit is None:
            self.circuit = tq.gates.Rz(angle=0, target=range(self.wavefunction.n_qubits)) # dummy empty circuit
        return self.circuit
        
    def get_wavefunction(self, variables=None):
        """
        Get the wavefunction from the circuit or the wavefunction
        """
        if self.circuit is not None and self.wavefunction is None:
            if variables is None:
                variables = self.variables
            self.wavefunction = tq.compile(self.circuit)(variables)
        return self.wavefunction

def compute_one_orb_rdm(mol:tqMolecule, circuit:QCircuit=None, variables:Variables=None, initial_state:QubitWaveFunction=None, one_orb:int=0)->np.ndarray:

    # Initialize the input state in a wrapper class
    state = Input_State(circuit=circuit, variables=variables, wavefunction=initial_state)
    if initial_state is not None and circuit is None:
        circuit = state.get_circuit()
        initial_state = state.get_wavefunction()
    elif circuit is None and initial_state is None:
        raise ValueError("Either a circuit or a wavefunction must be provided")
    up = mol.transformation.up(one_orb)
    down = mol.transformation.down(one_orb)
    ops = {
        "vacuum": tq.paulis.I(),
        "a_pu_dag": mol.make_creation_op(up),
        "a_pu": mol.make_annihilation_op(up),
        "n_pu": mol.make_number_op(up),
        "a_pd_dag": mol.make_creation_op(down),
        "a_pd": mol.make_annihilation_op(down),
        "n_pd": mol.make_number_op(down)
        }
    
    row = [
        ["vacuum"], # |00>
        ["a_pd_dag"], # |0down>
        ["a_pu_dag"], # |up0>
        ["a_pu_dag","a_pd_dag"] # |updown>
    ]

    # Create column as the daggered row
    column = []
    for tmp_ops in row:
        col_ops = []
        for op in tmp_ops:
            if op.endswith("_dag"):
                col_ops.append(op[:-4]) # removes the last 4 characters "_dag"
            else: 
                col_ops.append(op)
        # reverse as we are daggering
        col_ops.reverse()
        column.append(col_ops)

    rho = np.zeros((4,4))
    for i,state_i in enumerate(row):
        for j,state_j in enumerate(column):
            if i==j: # Select only specific terms as in https://iopscience.iop.org/article/10.1088/2058-9565/aca4ee/meta Eq.(27) 
                     # but the order is spin-down first and then spin-up
                
                # for convenience we multiply individual operators before
                op_i = tq.paulis.I()
                for v in state_i:
                    op_i *= ops[v]
                op_j = tq.paulis.I()
                for v in state_j:
                    op_j *= ops[v]

                P = op_i * (1-ops["n_pu"])*(1-ops["n_pd"]) * op_j

                if P.is_hermitian():
                    rho[i][j] = tq.simulate(tq.ExpectationValue(circuit,P), variables=variables, initial_state=initial_state)
                else:
                    P_herm, P_non_herm = P.split()
                    rho[i][j] = tq.simulate(tq.ExpectationValue(circuit,P_herm), variables=variables, initial_state=initial_state) + \
                        tq.simulate(tq.ExpectationValue(circuit,-1j * P_non_herm), variables=variables, initial_state=initial_state)

    return rho

def compute_two_orb_rdm(mol:tqMolecule, circuit:QCircuit=None, variables:Variables=None, initial_state:QubitWaveFunction=None, p_orb:int=0, q_orb:int=1, PSSR:bool=False, NSSR:bool=False)->np.ndarray:

    # Initialize the input state in a wrapper class
    state = Input_State(circuit=circuit, variables=variables, wavefunction=initial_state)
    if initial_state is not None and circuit is None:
        circuit = state.get_circuit()
        initial_state = state.get_wavefunction()
    elif circuit is None and initial_state is None:
        raise ValueError("Either a circuit or a wavefunction must be provided")
    
    pup = mol.transformation.up(p_orb)
    pdown = mol.transformation.down(p_orb)
    qup = mol.transformation.up(q_orb)
    qdown = mol.transformation.down(q_orb)

    ops = {
        "vacuum": tq.paulis.I(),
        "a_pu_dag": mol.make_creation_op(pup),
        "a_pu": mol.make_annihilation_op(pup),
        "n_pu": mol.make_number_op(pup),
        "a_pd_dag": mol.make_creation_op(pdown),
        "a_pd": mol.make_annihilation_op(pdown),
        "n_pd": mol.make_number_op(pdown),
        "a_qu_dag": mol.make_creation_op(qup),
        "a_qu": mol.make_annihilation_op(qup),
        "n_qu": mol.make_number_op(qup),
        "a_qd_dag": mol.make_creation_op(qdown),
        "a_qd": mol.make_annihilation_op(qdown),
        "n_qd": mol.make_number_op(qdown)
    }

    # For row and column order refer to https://pubs.acs.org/doi/10.1021/acs.jctc.0c00559 Figure 4
    row = [
        ["vacuum"], # |0000>
        ["a_qd_dag"], ["a_pd_dag"], # |000down>, |0down00>
        ["a_qu_dag"], ["a_pu_dag"], # |00up0>, |up000>
        ["a_pd_dag","a_qd_dag"], # |0down0down>
        ["a_qu_dag","a_qd_dag"], ["a_pd_dag","a_qu_dag"], ["a_pu_dag","a_qd_dag"], ["a_pu_dag","a_pd_dag"], # |00updown>, |0downup0>, |up00down>, |updown00>
        ["a_pu_dag","a_qu_dag"], # |up0up0>
        ["a_pd_dag","a_qu_dag","a_qd_dag"], ["a_pu_dag","a_pd_dag","a_qd_dag"], # |0downupdown>, |updown0down>
        ["a_pu_dag","a_qu_dag","a_qd_dag"], ["a_pu_dag","a_pd_dag","a_qu_dag"], # |up0updown>, |updownup0>
        ["a_pu_dag","a_pd_dag","a_qu_dag","a_qd_dag"] # |updownupdown>
    ]
    # Create column as the daggered row
    column = []
    for tmp_ops in row:
        col_ops = []
        for op in tmp_ops:
            if op.endswith("_dag"):
                col_ops.append(op[:-4]) # removes the last 4 characters "_dag"
            else: 
                col_ops.append(op)
        # reverse as we are daggering
        col_ops.reverse()
        column.append(col_ops)

    rho = np.zeros((16,16)) # 16x16 matrix
    for i,state_i in enumerate(row):
        for j,state_j in enumerate(column):
            if (i == j or 
                (i in {1, 2} and j in {1, 2}) or
                (i in {3, 4} and j in {3, 4}) or
                (i in {6, 7, 8, 9} and j in {6, 7, 8, 9}) or
                (i in {11, 12} and j in {11, 12}) or
                (i in {13, 14} and j in {13, 14})): # Select only specific terms as in https://pubs.acs.org/doi/10.1021/acs.jctc.0c00559 Figure 4

                # In order to evaluate SSR we make a local number counter for p and q on state i and state j
                N_p_i = N_q_i = N_p_j = N_q_j = 0
                for v,w in zip(state_i,state_j):
                    if v.startswith("a_p"):
                        N_p_i += 1
                    elif v.startswith("a_q"):
                        N_q_i += 1
                    if w.startswith("a_p"):
                        N_p_j += 1
                    elif w.startswith("a_q"):
                        N_q_j += 1
                # P-SSR (odd or even number of particles on local orbitals)
                if PSSR and (N_p_i%2!=N_p_j%2 or N_q_i%2!=N_q_j%2):
                    continue
                # N-SSR (same number of particles on local orbitals)
                if NSSR and (N_p_i!=N_p_j or N_q_i!=N_q_j):
                    continue

                # for convenience we multiply individual operators before
                op_i = tq.paulis.I()
                for v in state_i:
                    op_i *= ops[v]
                op_j = tq.paulis.I()
                for v in state_j:
                    op_j *= ops[v]

                P = op_i * (1-ops["n_pu"])*(1-ops["n_pd"])*(1-ops["n_qu"])*(1-ops["n_qd"]) * op_j
                if P.is_hermitian():
                    rho[i][j] = tq.simulate(tq.ExpectationValue(circuit,P), variables=variables, initial_state=initial_state)
                else:
                    P_herm, P_non_herm = P.split()
                    rho[i][j] = tq.simulate(tq.ExpectationValue(circuit,P_herm), variables=variables, initial_state=initial_state) + \
                        tq.simulate(tq.ExpectationValue(circuit,-1j * P_non_herm), variables=variables, initial_state=initial_state)

    return rho

def outer(psi):
    """Helper: Computes |psi><psi|."""
    return np.outer(psi, psi.conj())

def normalize_vector(psi):
    """Helper: Normalizes a vector if needed."""
    norm = np.linalg.norm(psi)
    if norm > 1e-9 and not np.isclose(norm, 1.0):
        return psi / norm
    return psi

def create_product_state(psi_A, psi_B):
    """
    Returns the density matrix for a pure product state:
    rho = |psi_A><psi_A| (tensor) |psi_B><psi_B|
    """
    # Normalize inputs
    psi_A = normalize_vector(psi_A)
    psi_B = normalize_vector(psi_B)
    
    rho_A = outer(psi_A)
    rho_B = outer(psi_B)
    return np.kron(rho_A, rho_B)

def create_separable_mixed_state(components, reorder_indices=None):
    """
    Returns a Separable Mixed State (convex sum of products).
    
    Args:
        components: List of tuples (probability, psi_A, psi_B)
        reorder_indices: (Optional) List of indices to permute the final matrix.
        
    Formula:
        rho = Sum( p_i * (|psiA_i><psiA_i| tensor |psiB_i><psiB_i|) )
    """
    if not components:
        raise ValueError("Components list is empty.")

    # 1. Normalize Probabilities
    probs = [c[0] for c in components]
    total_prob = sum(probs)
    
    if total_prob <= 0:
        raise ValueError("Sum of probabilities must be positive.")
        
    if not np.isclose(total_prob, 1.0):
        # Create a new list with normalized probabilities to avoid modifying input in place
        components = [(p / total_prob, psi_A, psi_B) for p, psi_A, psi_B in components]

    # Infer dimension from first element to initialize matrix
    _, pA_0, pB_0 = components[0]
    dim_A = len(pA_0)
    dim_B = len(pB_0)
    total_dim = dim_A * dim_B
    
    rho_total = np.zeros((total_dim, total_dim))
    
    for p, psi_A, psi_B in components:
        # 2. Normalize State Vectors
        psi_A = normalize_vector(psi_A)
        psi_B = normalize_vector(psi_B)
        
        # Create local density matrices
        rho_A = outer(psi_A)
        rho_B = outer(psi_B)
        
        # Tensor them together
        product_rho = np.kron(rho_A, rho_B)
        
        # Add weighted term
        rho_total += p * product_rho
    
    # Apply permutation if requested
    if reorder_indices is not None:
        if len(reorder_indices) != total_dim:
            raise ValueError(f"Permutation order length ({len(reorder_indices)}) "
                             f"must match matrix dimension ({total_dim}).")
        # Apply np.ix_ to reorder rows and columns simultaneously
        rho_total = rho_total[np.ix_(reorder_indices, reorder_indices)]
        
    return rho_total

def create_general_mixed_state(components, reorder_indices=None):
    """
    Returns a General Mixed State (sum of global pure states).
    This allows for mixing Entangled states.
    
    Args:
        components: List of tuples (probability, psi_AB)
        reorder_indices: (Optional) List of indices to permute the final matrix.
                         Useful for changing basis (e.g., to particle number basis).
        Note: psi_AB is a vector in the FULL Hilbert space.
        
    Formula:
        rho = Sum( p_i * |psiAB_i><psiAB_i| )
    """
    if not components:
        raise ValueError("Components list is empty.")

    # 1. Normalize Probabilities
    probs = [c[0] for c in components]
    total_prob = sum(probs)
    
    if total_prob <= 0:
        raise ValueError("Sum of probabilities must be positive.")
        
    if not np.isclose(total_prob, 1.0):
        components = [(p / total_prob, psi_AB) for p, psi_AB in components]

    # Infer dimension
    _, psi_0 = components[0]
    dim_total = len(psi_0)
    
    rho_total = np.zeros((dim_total, dim_total))
    
    for p, psi_AB in components:
        # 2. Normalize State Vector
        psi_AB = normalize_vector(psi_AB)
        
        # Create global density matrix term
        rho_term = outer(psi_AB)
        
        rho_total += p * rho_term
    
    # Apply permutation if requested
    if reorder_indices is not None:
        if len(reorder_indices) != dim_total:
            raise ValueError(f"Permutation order length ({len(reorder_indices)}) "
                             f"must match matrix dimension ({dim_total}).")
        # Apply np.ix_ to reorder rows and columns simultaneously
        rho_total = rho_total[np.ix_(reorder_indices, reorder_indices)]
        
    return rho_total

def change_basis(rho, direction='to_molecular'):
    """
    Switches matrix between Computational Kronecker Basis and Molecular Particle-Number Basis.
    
    Args:
        rho: The density matrix (16x16)
        direction: 'to_molecular' (Computational -> Molecular)
                   'to_computational'  (Molecular -> Computational)
    """
    MOLECULAR_PERM_ORDER = [0, 1, 4, 2, 8, 5, 3, 6, 9, 12, 10, 7, 13, 11, 14, 15]
    if direction == 'to_molecular':
        idx = MOLECULAR_PERM_ORDER
    elif direction == 'to_computational':
        idx = np.argsort(MOLECULAR_PERM_ORDER).tolist()
    else:
        raise ValueError("Direction must be 'to_molecular' or 'to_computational'")
        
    # Apply permutation to both rows and columns
    return rho[np.ix_(idx, idx)]

def get_reduced_states(rho, d_A, d_B):
    """Calculates reduced density matrices rho_A and rho_B."""
    # Reshape to (dim_A, dim_B, dim_A, dim_B)
    rho_tensor = rho.reshape(d_A, d_B, d_A, d_B)
    rho_A = np.trace(rho_tensor, axis1=1, axis2=3)
    rho_B = np.trace(rho_tensor, axis1=0, axis2=2)
    return rho_A, rho_B

def is_entangled_ppt(rho, d_A, d_B, tol=1e-9):
    """Peres-Horodecki Criterion (Partial Transpose)."""
    # Transpose B indices (1 and 3)
    rho_tensor = rho.reshape(d_A, d_B, d_A, d_B)
    rho_pt = rho_tensor.transpose(0, 3, 2, 1).reshape(d_A*d_B, d_A*d_B)
    min_eig = np.linalg.eigvalsh(rho_pt).min()
    return min_eig < -tol

def classify_7_categories(rho, tol=1e-7):
    rho = np.array(rho)
    dim = rho.shape[0]
    # Assume equal partition for molecular orbitals (e.g., 4x4 -> 2x2, 16x16 -> 4x4)
    d = int(np.sqrt(dim)) 
    
    # 1. Purity Check
    purity = np.trace(rho @ rho).real
    is_pure = np.abs(purity - 1.0) < tol
    
    rho_A, rho_B = get_reduced_states(rho, d, d)
    
    # --- BRANCH: PURE STATES ---
    if is_pure:
        # Check if subsystem is pure (Trace(rho_A^2) == 1)
        purity_A = np.trace(rho_A @ rho_A).real
        if np.abs(purity_A - 1.0) < tol:
            return "Pure Product State (Uncorrelated)"
        else:
            return "Pure Entangled State"

    # --- BRANCH: MIXED STATES ---
    else:
        # Check Entanglement
        if is_entangled_ppt(rho, d, d, tol):
            return "Mixed Entangled State"
        
        # If Separable:
        # Check for Product State (Uncorrelated)
        rho_product = np.kron(rho_A, rho_B)
        if np.linalg.norm(rho - rho_product) < tol:
            return "Mixed Product State (Uncorrelated Noise)"
            
        # Check Commutativity for Classical/Quantum correlations
        # [rho, rho_A x I]
        term_A = np.kron(rho_A, np.eye(d))
        comm_A = np.linalg.norm(rho @ term_A - term_A @ rho) < tol
        
        # [rho, I x rho_B]
        term_B = np.kron(np.eye(d), rho_B)
        comm_B = np.linalg.norm(rho @ term_B - term_B @ rho) < tol
        
        if comm_A and comm_B:
            return "Classically Correlated Separable State (CC)"
        elif comm_A or comm_B:
            return "Classical-Quantum State (CQ)"
        else:
            return "Discordant State (Separable but Quantum)"

def classify_hierarchy(rho, tol=1e-7):
    """
    Classifies according to the nested sets in Figure 1:
    D0 (Black) subset D_cl (Red) subset D_sep (Blue) subset All
    """
    rho = np.array(rho)
    d = int(np.sqrt(rho.shape[0]))
    
    rho_A, rho_B = get_reduced_states(rho, d, d)

    # 1. Check Entanglement (Outside D_sep)
    # The Gray Zone
    if is_entangled_ppt(rho, d, d, tol):
        return "Entangled (Outside D_sep)"

    # 2. Check Uncorrelated (D_0) - The Core
    # The Black Line
    rho_product = np.kron(rho_A, rho_B)
    if np.linalg.norm(rho - rho_product) < tol:
        return "D_0 (Uncorrelated / Product)"

    # 3. Check Classically Correlated (D_cl) - The Red Zone
    # Definition: Diagonal in the product basis of its marginals.
    # Mathematical test: Commutes with local reduced states.
    term_A = np.kron(rho_A, np.eye(d))
    term_B = np.kron(np.eye(d), rho_B)
    
    comm_A = np.linalg.norm(rho @ term_A - term_A @ rho) < tol
    comm_B = np.linalg.norm(rho @ term_B - term_B @ rho) < tol
    
    if comm_A and comm_B:
        return "D_cl (Classically Correlated - Red Zone)"

    # 4. Separable but Quantum (D_sep \ D_cl) - The Blue Zone
    return "D_sep (Separable with Quantum Discord - Blue Zone)"


from tequila import TequilaException
from .fermionic_base import FerMolecule
from .hybrid_base import HyMolecule
from tequila import Molecule as tqMolecule
from tequila.quantumchemistry.qc_base import QuantumChemistryBase

def Molecule(geometry: str = None,basis_set: str = None,nature: str = 'tequila',orbital_type: str = None,backend: str = None,guess_wfn=None,name: str = None,*args,**kwargs)->QuantumChemistryBase:
    """

    Parameters
    ----------
    geometry
        molecular geometry as string or as filename (needs to be in xyz format with .xyz ending)
    basis_set
        quantum chemistry basis set (sto-3g, cc-pvdz, etc)
    nature
        Molecule type (tequila, hybrid, fermionic)
    backend
        quantum chemistry backend (psi4, pyscf)
    guess_wfn
        pass down a psi4 guess wavefunction to start the scf cycle from
        can also be a filename leading to a stored wavefunction
    name
        name of the molecule, if not given it's auto-deduced from the geometry
        can also be done vice versa (i.e. geometry is then auto-deduced to name.xyz)
    args
    kwargs

    Returns
    -------
        Molecule object
    """
    if 'units' not in kwargs:
        kwargs['units'] = 'angstrom'
    if nature.lower()=='tequila' or nature.lower()=='t':
        return tqMolecule(geometry=geometry,basis_set=basis_set,orbital_type=orbital_type,backend=backend,guess_wfn=guess_wfn,name=name,*args,**kwargs)
    elif nature.lower()=='hybrid' or nature.lower()=='h':
        if 'select' in kwargs:
            select = kwargs['select']
            kwargs.pop('select')
        else: select = {}
        return HyMolecule(geometry=geometry,select=select,basis_set=basis_set,orbital_type=orbital_type,backend=backend,guess_wfn=guess_wfn,name=name,*args,**kwargs)
    elif nature.lower()=='fermionic' or nature.lower()=='f':
        return FerMolecule(geometry=geometry,basis_set=basis_set,orbital_type=orbital_type,backend=backend,guess_wfn=guess_wfn,name=name,*args,**kwargs)
    else:
        raise TequilaException(f"Molecule Nature not identified: {nature}. Only tequila, hybrid and fermionic allowed.")




import numpy as np
from scipy.spatial.transform import Rotation as R
from scipy.linalg import orthogonal_procrustes

class VectorUtils:
    # Unused for now
    @staticmethod
    def roughly_coplanar(vectors, threshold=1e-5):
        """
        Check if a list of vectors are roughly in the same plane, within a given threshold.

        :param vectors: A list of vectors, where each vector is a list/array of length 3.
                        Example: vectors = [[x1, y1, z1], [x2, y2, z2], ...]
        :param threshold: Acceptable deviation from complete coplanarity (default is 1e-5).
        :return: True if all vectors are roughly coplanar, False otherwise.
        """
        if len(vectors) < 3:
            # Fewer than 3 vectors are always coplanar by definition.
            return True

        vectors = [np.array(v) for v in vectors]

        # Compute the normal vector of the first two vectors via cross product
        normal = np.cross(vectors[0], vectors[1])

        # If the normal vector is zero (parallel vectors), return False unless all vectors are collinear
        if np.linalg.norm(normal) == 0:
            # Check if all vectors are collinear instead of coplanar
            for i in range(2, len(vectors)):
                # Cross product should still be near zero with collinear vectors
                if np.linalg.norm(np.cross(vectors[0], vectors[i])) >= threshold:
                    return False
            return True

        # Check subsequent vectors to see if they are in the same plane
        for i in range(2, len(vectors)):
            # Compute the normal vector for the plane formed by one of the original vectors and a new vector
            new_normal = np.cross(vectors[0], vectors[i])

            # Measure the deviation between the original plane normal and the new normal
            # Specifically, take the norm of the new cross-product and ensure it's near-zero
            if np.linalg.norm(new_normal) == 0:
                continue  # perfectly coplanar with this vector
            angle_difference = np.dot(normal, new_normal) / (np.linalg.norm(normal) * np.linalg.norm(new_normal))

            # If the angle difference is significant (outside threshold), their normals are not aligned -> non-coplanar
            # Clamp the value within the valid range of arccos (-1 to 1) to avoid numerical issues
            angle_difference = np.clip(angle_difference, -1.0, 1.0)

            # Use cosine similarity to ensure they are nearly parallel
            angle = np.arccos(angle_difference)

            # When the angle is too large (meaning normals differ significantly), the vectors are not coplanar
            if abs(angle) > threshold:
                return False

        return True
class MatrixUtils:
    @staticmethod
    def is_unitary(m):
        return np.allclose(np.eye(m.shape[0]), np.dot(m.T, m))

    @staticmethod
    def givens_rotation_matrix(size, i, j, theta):
        """Constructs a Givens rotation matrix that rotates between rows/columns i and j."""
        G = np.eye(size)
        G[i, i] = np.cos(theta)
        G[j, j] = np.cos(theta)
        G[i, j] = -np.sin(theta)
        G[j, i] = np.sin(theta)
        return G

    @staticmethod
    def gram_schmidt(vectors, tolerance=1e-8):
        """
        Modified Gram-Schmidt process to orthogonalize vectors, handling collinear cases.

        :param vectors: List of input vectors (Nx3).
        :param tolerance: Tolerance for determining zero vectors after projection.
        :return: Orthogonalized vectors, guaranteed to have the same number of rows as input.
        """
        orthogonal = []
        for i, v in enumerate(vectors):
            # Subtract projections onto previous orthogonal vectors
            for u in orthogonal:
                v -= np.dot(u, v) * u  # Subtract the component of v along u

            norm_v = np.linalg.norm(v)
            if norm_v < tolerance:
                # Handle collinear case: If two vectors are collinear, restore the second as the opposite
                v = -orthogonal[0]  # Flip the direction of the first vector
                norm_v = np.linalg.norm(v)

            orthogonal.append(v / norm_v)  # Normalize the vector

        return np.array(orthogonal)

    @staticmethod
    def kabsch_algorithm(P, Q):
        """
        Computes the optimal rotation matrix to align point clouds P and Q using the Kabsch algorithm.

        :param P: Original points (Nx3 array)
        :param Q: Target points (Nx3 array)
        :return: Rotation matrix R (3x3)
        """
        # Center the point clouds at the origin
        P_mean = np.mean(P, axis=0)
        Q_mean = np.mean(Q, axis=0)
        P_centered = P - P_mean
        Q_centered = Q - Q_mean

        # Compute the covariance matrix
        C = np.dot(P_centered.T, Q_centered)

        # Singular value decomposition (SVD) of the covariance matrix
        U, S, Vt = np.linalg.svd(C)

        # Compute the rotation matrix (enforcing proper rotation if necessary)
        R = np.dot(U, Vt)
        if np.linalg.det(R) < 0:
            U[:, -1] *= -1  # Flip the last column of U
            R = np.dot(U, Vt)

        return R

    @staticmethod
    def kabsch_algorithm_with_sign_correction(P, Q):
        """
        Extended Kabsch Algorithm to include sign correction.
        :param P: Source point set (Nx3).
        :param Q: Target point set (Nx3).
        :return: Rotation matrix R.
        """
        H = np.dot(P.T, Q)  # Covariance matrix
        U, S, Vt = np.linalg.svd(H)
        d = np.linalg.det(np.dot(Vt.T, U.T))  # Determinant of rotation matrix
        D = np.eye(3)  # Identity matrix
        if d < 0:
            D[2, 2] = -1  # Correct for reflection
        R = np.dot(np.dot(Vt.T, D), U.T)  # Adjusted rotation matrix
        return R

    @staticmethod
    def align_with_quaternions(P, Q):
        """
        Aligns vector sets P and Q using quaternions to find the optimal rotation matrix.

        :param P: Source vectors (Nx3 array)
        :param Q: Target vectors (Nx3 array)
        :return: Rotation matrix R (3x3)
        """
        assert len(P) == len(Q), "Both vector sets must have the same number of elements"

        # Find the quaternion that best aligns P to Q
        rotation, rmsd = R.align_vectors(Q, P)
        return rotation.as_matrix()

    @staticmethod
    def loewdin_orthogonalization(C, S):
        """
        Apply Loewdin orthogonalization to ensure the matrix C is unitary in the basis defined by overlap matrix S.

        Parameters:
        ----------
        C : ndarray
            Initial (non-unitary) guess matrix.
        S : ndarray
            Overlap matrix from the molecular basis.

        Returns:
        -------
        C_orth : ndarray
            Orthonormalized coefficient matrix.
        """
        # Eigen decomposition of S
        sv, U = np.linalg.eigh(S)

        # Construct S^(-1/2)
        S_inv_sqrt = U @ np.diag(1.0 / np.sqrt(sv)) @ U.T

        # Apply transformation to orthogonalize C
        C_orth = S_inv_sqrt @ C

        return C_orth

    @staticmethod
    def make_unitary(C):
        """
        Ensure that a matrix C is unitary by performing QR decomposition.

        Parameters:
        ----------
        C : ndarray
            Input (possibly non-unitary) matrix.

        Returns:
        -------
        Q : ndarray
            Strictly unitary matrix.
        """
        Q, R = np.linalg.qr(C)
        return Q
class HybridizationUtils:
    @staticmethod
    def procrustes_alignment(align_source, align_target, gram_correction=True):
        """
        Aligns vectors using the orthogonal Procrustes algorithm.
        :param align_source: Source vector set (Nx3).
        :param align_target: Target vector set (Nx3).
        :param gram_correction: Apply Gram-Schmidt orthogonalization after alignment.
        :return: Rotated source vector set.
        """
        R, _ = orthogonal_procrustes(
            align_source,
            align_target
        )

        # Rotate the px, py, pz components of the hybrid orbitals
        rotated = np.dot(
            align_source,
            R.T
        )

        if gram_correction: rotated = MatrixUtils.gram_schmidt(rotated)  # Re-orthogonalize/normalize directions
        return rotated

    @staticmethod
    def kabsch_alignment(align_source, align_target, gram_correction=False):
        """
        Aligns vectors using the Kabsch algorithm (pure rotation).
        :param align_source: Source vector set (Nx3).
        :param align_target: Target vector set (Nx3).
        :param gram_correction: Apply Gram-Schmidt orthogonalization after alignment.
        :return: Rotated source vector set.
        """
        R = MatrixUtils.kabsch_algorithm_with_sign_correction(align_source, align_target)  # Get pure rotation matrix
        rotated = np.dot(align_source, R.T)  # Apply transformation

        if gram_correction:
            rotated = MatrixUtils.gram_schmidt(rotated)  # Re-orthogonalize/normalize directions

        return rotated

    @staticmethod
    def quaternion_alignment(align_source, align_target, gram_correction=True):
        """
        Aligns vectors using quaternions (via scipy's R.align_vectors).
        :param align_source: Source vector set (Nx3).
        :param align_target: Target vector set (Nx3).
        :param gram_correction: Apply Gram-Schmidt orthogonalization after alignment.
        :return: Rotated source vector set.
        """
        assert len(align_source) == len(align_target), "Source and target must have the same number of points."

        # Find the rotation matrix using scipy's quaternion-based vector alignment
        rotation, rmsd = R.align_vectors(align_target, align_source)  # align_vectors aligns src -> tgt
        R_matrix = rotation.as_matrix()

        # Rotate the source vectors
        rotated = np.dot(align_source, R_matrix.T)

        if gram_correction:
            rotated = MatrixUtils.gram_schmidt(rotated)  # Re-orthogonalize/normalize directions

        return rotated

    def correct_signs(rotated, target):
        # Loop over each pair of vectors
        for i in range(len(rotated)):
            if np.dot(rotated[i], target[i]) < 0:  # Check if signs are flipped
                rotated[i] *= -1  # Flip the sign of the rotated vector
        return rotated
class AtomUtils:
    max_electrons_by_shell = [2, 8, 18, 32, 32, 18, 8]
class XYZUtils:
    def cleanup(xyz_string):
        lines = [line.strip() for line in xyz_string.strip().splitlines()]
        for i, line in enumerate(lines):
            if len(line) != 0:  # Found the first line with content
                if ' ' in line:
                    # Join the remaining lines
                    return '\n'.join(lines[i:])
                else:  # Starts with number and comment lines
                    # Remove the first 2 lines with content, and join the remaining lines
                    return '\n'.join(lines[i + 2:])
        return ''  # If no lines with spaces exist, return an empty string


from pyscf import gto, scf
from pyscf.tools import cubegen
import tequila as tq
from tequila.quantumchemistry.qc_base import QuantumChemistryBase
from tequila import TequilaException
from sunrise.miscellaneous.bar import giuseppe_bar
import sys

def plot_MO(molecule:QuantumChemistryBase=None,filename:str=None,orbital:list=None,print_orbital:bool=True,density:bool=False,mep:bool=False):
    """
    Small function to save the MOs into Cube files
    Parameters
    ----------
    filename : Cube file will be saved as name+orb_index
    orbital: index of the orbitals to save
    molecule: molecule to plot the orbitals from
    print_orbital: whether to print the MOs
    density: whether to print the electron density
    mep: whether to plot the molecular electrostatic potential
    """
    if molecule is None:
        raise TequilaException("No Molecule to save orbitals from")
    if filename is None:
        filename = molecule.parameters.name +'-'+ molecule.integral_manager._basis_name+'-'+molecule.integral_manager._orbital_type
    if orbital is None:
        orbital = [i.idx_total for i in molecule.integral_manager.orbitals]
    pmol = gto.Mole()
    pmol.build(atom=molecule.parameters.geometry, basis=molecule.parameters.basis_set, spin=0)
    if density or mep:
        mf = scf.RHF(pmol).run()
        mf.mo_coeff=molecule.integral_manager.orbital_coefficients
    if print_orbital:
        for i,idx in enumerate(orbital):
            giuseppe_bar(step=i,total_steps=len(orbital))
            cubegen.orbital(pmol,  str(idx)+"_"+filename+"_MO.cube", molecule.integral_manager.orbital_coefficients[:, idx])
        giuseppe_bar(step=i+1,total_steps=len(orbital))
        sys.stdout.write('\n')
        sys.stdout.flush()
    if density:
        cubegen.density(pmol, filename + '_density.cube', mf.make_rdm1())
    if mep:
        cubegen.mep(pmol, filename + '_mep.cube', mf.make_rdm1())



from tequila.optimizers import minimize as tminimize
import typing
from tequila.circuit.compiler import CircuitCompiler
from tequila.objective.objective import Objective,ExpectationValueImpl,Variable,assign_variable,identity,FixedVariable
from tequila.circuit.noise import NoiseModel
from tequila import TequilaException,QCircuit,QubitWaveFunction,Molecule,TequilaWarning
from tequila.circuit.gradient import grad as tgrad
from tequila.objective import QTensor,format_variable_dictionary
from tequila.simulators.simulator_api import compile
import typing
from numpy import vectorize,zeros
from tequila.autograd_imports import jax, __AUTOGRAD__BACKEND__
from typing import Dict, Union, Hashable,Callable
from numbers import Number
from numbers import Real as RealNumber
from ..fermionic_operations import FCircuit
from sunrise.expval.pyscf_molecule import MoleculeFromPyscf 
from pyscf.gto import Mole
import warnings

def minimize(objective,method: str = "bfgs",variables: list = None,initial_values: Union[dict, Number, Callable] = 0.0,maxiter: int = None,silent:bool=True,*args,**kwargs):
    if type(objective).__name__ == 'TequilaBraket':
        return tminimize(objective=objective.build(),method=method,variables=variables,initial_values=initial_values,maxiter=maxiter,silent=silent,args=args,kwargs=kwargs)
    if type(objective).__name__ in  ['TCCBraket','FQEBraKet']:
        objective = Objective([objective])
    if any([type(arg).__name__ in  ['TCCBraket','FQEBraKet'] for arg in objective.args]):
        dE = grad(objective=objective,variable=variables,args=args,kwargs=kwargs)
        return tminimize(objective=objective,gradient=dE,method=method,variables=variables,initial_values=initial_values,maxiter=maxiter,silent=silent,args=args,kwargs=kwargs)
    else:
        return tminimize(objective=objective,method=method,variables=variables,initial_values=initial_values,maxiter=maxiter,silent=silent,args=args,kwargs=kwargs)

#FIXME: Placeholder functions while not commited to tequila
def grad(objective: Union[Objective, QTensor], variable: Variable = None, no_compile=False, *args, **kwargs):
    """
    wrapper function for getting the gradients of Objectives,ExpectationValues, Unitaries (including single gates), and Transforms.
    :param obj (QCircuit,ParametrizedGateImpl,Objective,ExpectationValue,Transform,Variable): structure to be differentiated
    :param variables (list of Variable): parameter with respect to which obj should be differentiated.
        default None: total gradient.
    return: dictionary of Objectives, if called on gate, circuit, exp.value, or objective; if Variable or Transform, returns number.
    """
    if type(objective).__name__ in  ['TCCBraket','FQEBraKet']:
        objective = Objective([objective],transformation=identity)
    elif type(objective).__name__ == 'TequilaBraket':
        objective = objective.build()
    if variable is None:
        # None means that all components are created
        variables = objective.extract_variables()
        result = {}

        if len(variables) == 0:
            raise TequilaException("Error in gradient: Objective has no variables")
        for k in variables:
            assert k is not None
            result[k] = grad(objective, k, no_compile=no_compile)
        return result
    elif isinstance(variable,list):
        result = {}
        if len(variable) == 0:
            raise TequilaException("Error in gradient: Objective has no variables")

        for k in variable:
            assert k is not None
            result[k] = grad(objective, k, no_compile=no_compile)
        return result
    else:
        variable = assign_variable(variable)

    if isinstance(objective, QTensor):
        f = lambda x: grad(objective=x, variable=variable, *args, **kwargs)
        ff = vectorize(f)
        return ff(objective)

    if variable not in objective.extract_variables():
        return Objective()

    # objective translation
    # if the objective was already translated to a backend
    # we need to reverse that here
    our = False
    if any([type(arg).__name__ in  ['TCCBraket','FQEBraKet'] for arg in objective.args]):
        our = True
        no_compile = True
    elif not our and objective.is_translated():
        raise TequilaException(
            "\n\ngradient of:{}\ncan not form gradient that was already compiled to a quantum backend\ntq.grad neds to be applied to the abstract - non compiled objective\nE.g. for the (compiled) objective E1 \n\tE1 = tq.compile(E0)\ninstead of doing\n\tdE = tq.grad(E1)\ndo\n\tdE = tq.grad(E0)\nand compile dE afterwards (if wanted) with\n\tdE = tq.compile(dE)\n".format(
                str(objective)
            )
        )

    # circuit compilation
    if no_compile:
        compiled = objective
    else:
        compiler = CircuitCompiler(
            multitarget=True,
            trotterized=True,
            hadamard_power=True,
            power=True,
            controlled_phase=True,
            controlled_rotation=True,
            gradient_mode=True,
        )

        compiled = compiler(objective, variables=[variable])

    if variable not in compiled.extract_variables():
        raise TequilaException("Error in taking gradient. Objective does not depend on variable {} ".format(variable))

    if isinstance(objective, ExpectationValueImpl):
        return __grad_expectationvalue(E=objective, variable=variable)
    elif objective.is_expectationvalue():
        return __grad_expectationvalue(E=compiled.args[-1], variable=variable)
    elif isinstance(compiled, Objective) or (hasattr(compiled, "args") and hasattr(compiled, "transformation")):
        return __grad_objective(objective=compiled, variable=variable)
    else:
        raise TequilaException("Gradient not implemented for other types than ExpectationValue and Objective.")

def __grad_objective(objective: Objective, variable: Variable):
    args = objective.args
    transformation = objective.transformation
    dO = None
    processed_expectationvalues = {}
    for i, arg in enumerate(args):
        if __AUTOGRAD__BACKEND__ == "jax":
            df = jax.grad(transformation, argnums=i)
        elif __AUTOGRAD__BACKEND__ == "autograd":
            df = jax.grad(transformation, argnum=i)
        else:
            raise TequilaException("Can't differentiate without autograd or jax")

        # We can detect one simple case where the outer derivative is const=1
        if transformation is None or transformation == identity:
            outer = 1.0
        else:
            outer = Objective(args=args, transformation=df)

        if hasattr(arg, "U"):
            # save redundancies
            if arg in processed_expectationvalues:
                inner = processed_expectationvalues[arg]
            else:
                inner = __grad_inner(arg=arg, variable=variable)
                processed_expectationvalues[arg] = inner
        else:
            # this means this inner derivative is purely variable dependent
            inner = __grad_inner(arg=arg, variable=variable)

        if inner == 0.0:
            # don't pile up zero expectationvalues
            continue

        if dO is None:
            dO = outer * inner
        else:
            dO = dO + outer * inner

    if dO is None:
        raise TequilaException("caught None in __grad_objective")
    return dO

def __grad_inner(arg, variable):
    """
    a modified loop over __grad_objective, which gets derivatives
    all the way down to variables, return 1 or 0 when a variable is (isnt) identical to var.
    :param arg: a transform or variable object, to be differentiated
    :param variable: the Variable with respect to which par should be differentiated.
    :ivar var: the string representation of variable
    """

    assert isinstance(variable, Variable)
    if isinstance(arg, Variable):
        if arg == variable:
            return 1.0
        else:
            return 0.0
    elif isinstance(arg, FixedVariable):
        return 0.0
    elif isinstance(arg, ExpectationValueImpl):
        return __grad_expectationvalue(arg, variable=variable)
    elif hasattr(arg, "abstract_expectationvalue"):
        E = arg.abstract_expectationvalue
        dE = __grad_expectationvalue(E, variable=variable)
        return compile(dE, **arg._input_args)
    elif hasattr(arg,'grad'):
        return arg.grad(variable)
    else:
        return __grad_objective(objective=arg, variable=variable)

def __grad_expectationvalue(E: ExpectationValueImpl, variable: Variable):
    """
    implements the analytic partial derivative of a unitary as it would appear in an expectation value. See the paper.
    :param unitary: the unitary whose gradient should be obtained
    :param variables (list, dict, str): the variables with respect to which differentiation should be performed.
    :return: vector (as dict) of dU/dpi as Objective (without hamiltonian)
    """

    hamiltonian = E.H
    unitary = E.U
    if not (unitary.verify()):
        raise TequilaException("error in grad_expectationvalue unitary is {}".format(unitary))

    # fast return if possible
    if variable not in unitary.extract_variables():
        return 0.0

    param_gates = unitary._parameter_map[variable]

    dO = Objective()
    for idx_g in param_gates:
        idx, g = idx_g
        dOinc = __grad_shift_rule(unitary, g, idx, variable, hamiltonian)
        dO += dOinc

    assert dO is not None
    return dO

def __grad_shift_rule(unitary, g, i, variable, hamiltonian):
    """
    function for getting the gradients of directly differentiable gates. Expects precompiled circuits.
    :param unitary: QCircuit: the QCircuit object containing the gate to be differentiated
    :param g: a parametrized: the gate being differentiated
    :param i: Int: the position in unitary at which g appears
    :param variable: Variable or String: the variable with respect to which gate g is being differentiated
    :param hamiltonian: the hamiltonian with respect to which unitary is to be measured, in the case that unitary
        is contained within an ExpectationValue
    :return: an Objective, whose calculation yields the gradient of g w.r.t variable
    """

    # possibility for overwride in custom gate construction
    if hasattr(g, "shifted_gates"):
        inner_grad = __grad_inner(g.parameter, variable)
        shifted = g.shifted_gates()
        dOinc = Objective()
        for x in shifted:
            w, g = x
            Ux = unitary.replace_gates(positions=[i], circuits=[g])
            wx = w * inner_grad
            Ex = Objective.ExpectationValue(U=Ux, H=hamiltonian)
            dOinc += wx * Ex
        return dOinc
    else:
        raise TequilaException("No shift found for gate {}\nWas the compiler called?".format(g))

#FIXME: When commited it should be replaced by:
# def grad(objective: Union[Objective, QTensor], variable: Variable = None, no_compile=False, *args, **kwargs):
#     """
#     wrapper function for getting the gradients of Objectives,ExpectationValues, Unitaries (including single gates), and Transforms.
#     :param obj (QCircuit,ParametrizedGateImpl,Objective,ExpectationValue,Transform,Variable): structure to be differentiated
#     :param variables (list of Variable): parameter with respect to which obj should be differentiated.
#         default None: total gradient.
#     return: dictionary of Objectives, if called on gate, circuit, exp.value, or objective; if Variable or Transform, returns number.
#     """

#     if any([type(arg).__name__ in  ['TCCBraket','FQEBraKet'] for arg in objective.args]):
#         return tgrad(objective=objective, variable = variable, no_compile=True, *args, **kwargs)
#     else:
#         return tgrad(objective=objective, variable=variable, no_compile=no_compile, *args, **kwargs)

def simulate(
    objective: typing.Union[FCircuit,"Objective", "QCircuit", "QTensor"],
    variables: Dict[Union[Variable, Hashable], RealNumber] = None,
    samples: int = None,
    backend: str = None,
    noise: NoiseModel = None,
    device: str = None,
    initial_state: Union[int, QubitWaveFunction] = 0,
    *args,
    **kwargs,
) -> Union[RealNumber, QubitWaveFunction]:
    """Simulate a tequila objective or circuit

    Parameters
    ----------
    objective: Objective:
        tequila objective or circuit
    variables: Dict:
        The variables of the objective given as dictionary
        with keys as tequila Variables/hashable types and values the corresponding real numbers
    samples : int, optional:
        if None a full wavefunction simulation is performed, otherwise a fixed number of samples is simulated
    backend : str, optional:
        specify the backend or give None for automatic assignment
    noise: NoiseModel, optional:
        specify a noise model to apply to simulation/sampling
    device:
        a device upon which (or in emulation of which) to sample
    initial_state: int or QubitWaveFunction:
        the initial state of the circuit
    *args :

    **kwargs :
        read_out_qubits = list[int] (define the qubits which shall be measured, has only effect on pure QCircuit simulation with samples)

    Returns
    -------
    float or QubitWaveFunction
        the result of simulation.
    """

    variables = format_variable_dictionary(variables)

    if variables is None and not (len(objective.extract_variables()) == 0):
        raise TequilaException(
            "You called simulate for a parametrized type but forgot to pass down the variables: {}".format(
                objective.extract_variables()
            )
        )
    
    if isinstance(objective,list):
        return [simulate(op,variables,samples,backend,noise,device,initial_state,*args,**kwargs) for op in objective]
    if type(objective).__name__ in ['TCCBraket','FQEBraKet']:
        return objective(variables=variables)
    if type(objective).__name__ == 'TequilaBraket':
        return simulate(objective=objective.build(),variables=variables,samples=samples,backend=backend,noise=noise,device=device,initial_state=initial_state,*args,**kwargs)
    if isinstance(objective,FCircuit):
        if initial_state != 0 :
            raise TequilaWarning("sun.Simulate(FCircuit) doesn't support the keyword initial_state, you have to provide it from\n" \
            "the FCicuit.initial_state, in UPTHENDOWN mandatory")
        mol = None
        nmo = None
        if 'molecule' in kwargs:
            mol = kwargs['molecule']
            kwargs.pop('molecule')
        elif 'mol' in kwargs:
            mol = kwargs['mol']
            kwargs.pop('mol')
        elif 'n_mo' in kwargs:
            nmo = kwargs['nmo']
            kwargs.pop('nmo')
        elif 'n_orbitals' in kwargs:
            nmo = kwargs['n_orbitals']
            kwargs.pop('n_orbitals')
        else:
            avisa = True
            nmo = (max(objective.qubits)//2) + 1
        if mol is not None:
            if isinstance(mol,Mole):
                mol = MoleculeFromPyscf(mol)
        elif nmo is not None:
            if 'transformation' in kwargs:
                transformation = kwargs['transformation']
                kwargs.pop('transformation')
            else: transformation = 'Jordan-Wigner'
            h = zeros((nmo,nmo))
            g = zeros((nmo,nmo,nmo,nmo))
            mol = Molecule(geometry= "\n".join(f"H 0.0 0.0 {i}" for i in range(nmo)),basis_set='sto-3g',one_body_integrals=h,two_body_integrals=g,transformation=transformation,units='Angstrom',nature='f',backend='pyscf')
            if avisa and not all([ not g.reordered for g in objective.gates]):
                warnings.warn("Some indices were provided in reordered, the output may not be correct",TequilaWarning)
        mol.transformation.up_then_down = True
        U = objective.to_qcircuit(mol)
        #TODO: This is a temporaly bypass while more sofisticate initial states
        return simulate(objective=U,variables=variables,samples=samples,backend=backend,noise=noise,device=device,args=args,kwargs=kwargs)
        #IDEA this is should be continued when fixed
        # ini = f''.join(f'{ini[i].real}|{bin(i)[2:]}>'.format() for i in main)
        # ini = QubitWaveFunction.from_string(f''.join(f'{ini[i].real}|{bin(i)[2:]}>'.format() for i in main))
        # U = objective.to_qcircuit(mol)
        # return  simulate(objective=U,initial_state=ini,variables=variables,samples=samples,backend=backend,noise=noise,device=device,args=args,kwargs=kwargs)
            
    compiled_objective = compile(
        objective=objective,
        samples=samples,
        variables=variables,
        backend=backend,
        noise=noise,
        device=device,
        *args,
        **kwargs,
    )

    return compiled_objective(variables=variables, samples=samples, initial_state=initial_state, *args, **kwargs)



from sunrise import FCircuit
from . import Braket
from tequila.quantumchemistry import optimize_orbitals as tq_opt_orbs
from tequila.quantumchemistry.orbital_optimizer import OptimizeOrbitalsResult
from ..molecules.fermionic_base.fer_base import FermionicBase
from ..molecules.hybrid_base.HybridBase import HybridBase
from .minimize import minimize


def optimize_orbitals(molecule,circuit=FCircuit,backend:str='tequila',pyscf_arguments=None,silent=False,backend_kwargs:dict=None,initial_guess=None,return_mcscf=False,
    molecule_factory=None,molecule_arguments=None,restrict_to_active_space=True,*args,**kwargs)->OptimizeOrbitalsResult:
    """

    Parameters
    ----------
    molecule: The tequila molecule whose orbitals are to be optimized
    circuit: The FCircuit that defines the ansatz to the wavefunction in the VQE
             can be None, if a customized vqe_solver is passed that can construct a circuit
    backend: Fermionic Backend, will be created a tequila.chemistry.optimize_orbitals with custom vqe_solver
    pyscf_arguments: Arguments for the MCSCF structure of PySCF, if None, the defaults are {"max_cycle_macro":10, "max_cycle_micro":3} (see here https://pyscf.org/pyscf_api_docs/pyscf.mcscf.html)
    silent: silence printout
    backend_kwargs: Optional arguments for a customized selected backed
                          for the default solver: backend_kwargs={"optimizer_arguments":A, "restrict_to_hcb":False} where A holds the kwargs for tq.minimize
                          restrict_to_hcb keyword controls if the standard (in whatever encoding the molecule structure has) Hamiltonian is constructed or the hardcore_boson hamiltonian
    initial_guess: Initial guess for the MCSCF module of PySCF (Matrix of orbital rotation coefficients)
                   The default (None) is a unit matrix
                   predefined commands are
                        initial_guess="random"
                        initial_guess="random_loc=X_scale=Y" with X and Y being floats
                        This initialized a random guess using numpy.random.normal(loc=X, scale=Y) with X=0.0 and Y=0.1 as defaults
    return_mcscf: return the PySCF MCSCF structure after optimization
    molecule_arguments: arguments to pass to molecule_factory or default molecule constructor | only change if you know what you are doing
    args: just here for convenience
    kwargs: just here for conveniece

    Returns
    -------
        Optimized Tequila Molecule
    """
    class solver:
        def __init__(self,backend:str='tequila',circuit:FCircuit=None):
            self.backend = backend
            self.U = circuit
        def __call__(self, H, circuit, molecule, **backend_kwargs):
            if 'silent' in backend_kwargs:
                silent = backend_kwargs['silent']
                backend_kwargs.pop('silent')
            else: silent = True
            return minimize(Braket(backend=backend,molecule=molecule,circuit=self.U,**backend_kwargs),silent=silent)
    vqe_solver = solver(backend=backend,circuit=circuit)
    
    if isinstance(molecule,HybridBase):
        if molecule_arguments is None:
            molecule_arguments = {"select": molecule.select, "condense": molecule.condense,
                                  "two_qubit": molecule.two_qubit,
                                  "integral_tresh": molecule.integral_tresh,"parameters": molecule.parameters,
                                  "transformation": molecule.transformation,"backend":'pyscf'}
        else:
            mol_args = {"select": molecule.select, "condense": molecule.condense,
                                  "two_qubit": molecule.two_qubit,
                                  "integral_tresh": molecule.integral_tresh,"parameters": molecule.parameters,
                                  "transformation": molecule.transformation,"backend":'pyscf'}
            mol_args.update(molecule_arguments)
            molecule_arguments = mol_args
            circuit = circuit.to_qcircuit(molecule=molecule)
        if molecule_factory is None:
            molecule_factory = HybridBase
        vqe_solver = None
    elif isinstance(molecule,FermionicBase):
        if molecule_factory is None:
            molecule_factory = FermionicBase
        molecule.fermionic_backend = backend
        if molecule_arguments is None:
            molecule_arguments = {'fermionic_backend':backend,'parameters':molecule.parameters}
        else: 
            molecule_arguments['fermionic_backend']=backend
            if 'parameters' not in molecule_arguments:
                molecule_arguments['parameters'] = molecule.parameters
    else:
        circuit.to_qcircuit(molecule=molecule)
        vqe_solver = None
    result = tq_opt_orbs(molecule=molecule,circuit=circuit,vqe_solver=vqe_solver,pyscf_arguments=pyscf_arguments,silent=silent,initial_guess=initial_guess,return_mcscf=return_mcscf,molecule_factory=molecule_factory,molecule_arguments=molecule_arguments,restrict_to_active_space=restrict_to_active_space,*args,**kwargs)
    if isinstance(molecule,HybridBase):
        result.molecule = HybridBase(**molecule_arguments, integral_manager=result.molecule.integral_manager)
    elif isinstance(molecule,FermionicBase):
        result.molecule = FermionicBase(parameters=result.molecule.parameters, integral_manager=result.molecule.integral_manager)
    return result


import typing
import warnings

from tequila import TequilaWarning
from .HybridBase import HybridBase
from tequila.quantumchemistry.chemistry_tools import ParametersQC

SUPPORTED_QCHEMISTRY_BACKENDS = ["base", "pyscf"] # , "psi4", "madness"
INSTALLED_QCHEMISTRY_BACKENDS = {"base": HybridBase} #"madness": QuantumChemistryMadness

try:
    import QuantumChemistryPsi4
    INSTALLED_QCHEMISTRY_BACKENDS["psi4"] = QuantumChemistryPsi4
except ImportError:
    pass

try:
    import pyscf
    from .pyscf_interface import PySCF
    INSTALLED_QCHEMISTRY_BACKENDS["pyscf"] = PySCF
except ImportError:
    pass


def HyMolecule(geometry: str = None,basis_set: str = None,transformation: typing.Union[str, typing.Callable] = None,
             orbital_type: str = None,backend: str = None,guess_wfn=None,name: str = None,*args,**kwargs) -> HybridBase:
    """

    Parameters
    ----------
    geometry
        molecular geometry as string or as filename (needs to be in xyz format with .xyz ending)
    basis_set
        quantum chemistry basis set (sto-3g, cc-pvdz, etc)
    transformation
        The Fermion to Qubit Transformation (jordan-wigner, bravyi-kitaev, bravyi-kitaev-tree and whatever OpenFermion supports)
    backend
        quantum chemistry backend (psi4, pyscf)
    guess_wfn
        pass down a psi4 guess wavefunction to start the scf cycle from
        can also be a filename leading to a stored wavefunction
    name
        name of the molecule, if not given it's auto-deduced from the geometry
        can also be done vice versa (i.e. geometry is then auto-deduced to name.xyz)
    args
    kwargs

    Returns
    -------
        The Fermion to Qubit Transformation (jordan-wigner, bravyi-kitaev, bravyi-kitaev-tree and whatever OpenFermion supports)
    """

    # failsafe for common mistake
    if "basis" in kwargs:
        warnings.warn(
            "called molecule with keyword \"basis={0}\" converting it to \"basis_set={0}\"".format(kwargs["basis"]),
            TequilaWarning)
        if basis_set is not None:
            warnings.warn("did not convert as \"basis_set={}\" was already given".format(basis_set), TequilaWarning)
        basis_set = kwargs["basis"]

    keyvals = {}
    for k, v in kwargs.items():
        if k in ParametersQC.__dict__.keys():
            keyvals[k] = v

    if "parameters" in kwargs:
        parameters = kwargs["parameters"]
        kwargs.pop("parameters")
    else:
        parameters = ParametersQC(name=name, geometry=geometry, basis_set=basis_set, multiplicity=1, **keyvals)

    integrals_provided = all([key in kwargs for key in ["one_body_integrals", "two_body_integrals"]])
    if integrals_provided and backend is None:
        backend = "base"
    #
    if backend is None:
        if basis_set is None or basis_set.lower() in ["madness", "mra", "pno"]:
            backend = "madness"
            basis_set = "mra"
            parameters.basis_set = basis_set
            if orbital_type is not None and orbital_type.lower() not in ["pno", "mra-pno"]:
                warnings.warn(
                    "only PNOs supported as orbital_type without basis set. Setting to pno - You gave={}".format(orbital_type), TequilaWarning)
            orbital_type = "pno"
        else:
            if orbital_type is not None and orbital_type.lower() not in ["hf", "native"]:
                warnings.warn(
                    "only hf and native supported as orbital_type with basis-set. Setting to hf - You gave={}".format(
                        orbital_type), TequilaWarning)
                orbital_type = "hf"
            if orbital_type is None:
                orbital_type = "hf"

            if "psi4" in INSTALLED_QCHEMISTRY_BACKENDS:
                backend = "psi4"
            elif "pyscf" in INSTALLED_QCHEMISTRY_BACKENDS:
                backend = "pyscf"
            else:
                raise Exception("No quantum chemistry backends installed on your system")

    elif backend == "base":
        if not integrals_provided:
            raise Exception("No quantum chemistry backends installed on your system\n"
                            "To use the base functionality you need to pass the following tensors via keyword\n"
                            "one_body_integrals, two_body_integrals\n")
        else:
            backend = "base"

    if backend not in SUPPORTED_QCHEMISTRY_BACKENDS:
        raise Exception(str(backend) + " is not (yet) supported by Hybrid Molecule")
    if backend not in INSTALLED_QCHEMISTRY_BACKENDS:
        raise Exception(str(backend) + " was not found on your system")

    if guess_wfn is not None and backend != 'psi4':
        raise Exception("guess_wfn only works for psi4")

    if basis_set is None and backend.lower() not in ["base", "madness"] and not integrals_provided:
        raise Exception("no basis_set or integrals provided for backend={}".format(backend))
    elif basis_set is None:
        basis_set = "custom"
        parameters.basis_set = basis_set

    return INSTALLED_QCHEMISTRY_BACKENDS[backend.lower()](parameters=parameters, transformation=transformation, orbital_type=orbital_type, guess_wfn=guess_wfn, *args, **kwargs)



import numbers
import typing
from tequila.objective.objective import FixedVariable,Variable
from .circuit import FCircuit
from tequila import TequilaException
from tequila import assign_variable
from copy import deepcopy

class FGateImpl:
    def __init__(self,indices:typing.Union[list,tuple],variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]=None,reordered:bool=False):
        self.reordered:bool=reordered
        self._indices:list = indices
        self.variables=variables
        self._name:str = 'GenericFermionic'
        self.verify()
        return FCircuit.wrap_gate(gate=self)

    @property
    def name(self):
        return self._name
    
    def is_controlled(self):
        return False #TODO: at some point we should

    def to_upthendown(self,norb:int):
        if not self.reordered:
            self._indices = [[(idx[0]//2+(idx[0]%2)*norb,idx[1]//2+(idx[1]%2)*norb) for idx in gate] for gate in self._indices]
        return self
    
    def to_udud(self,norb:int):
        if self.reordered:
            self._indices = [[(2*(idx[0]%norb)+(idx[0]>=norb),2*(idx[1]%norb)+(idx[1]>=norb)) for idx in gate] for gate in self._indices]
        return self
    
    def __str__(self):
        return f'{self.name}(indices = {self.indices} ,variables = {repr(self.variables)})'

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        if self._name != other._name:
            return False
        elif self._indices != other._indices:
            return False
        elif self.extract_variables() != other.extract_variables():
            return False
        return True
            
    def is_parameterized(self)->bool:
        return not isinstance(self.variables,(FixedVariable,numbers.Number))

    @property
    def indices(self):
        return self._indices
     
    @indices.setter
    def indices(self,indices):
        self._indices = indices
    
    def extract_variables(self)->list[Variable]:
        if self.is_parameterized() and hasattr(self.variables, "extract_variables"):
            return self.variables.extract_variables()
        else:
            return []

    @property
    def variables(self)->Variable:
        return self._variables
    
    @variables.setter
    def variables(self,variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]):
        self._variables:Variable = assign_variable(variable=variables)

    def verify(self):
        assert isinstance(self._variables,(typing.Hashable, numbers.Real, Variable, FixedVariable))
        if isinstance(self._indices[0],numbers.Number): #[1,3]
            self._indices = [[tuple(self._indices),],] #->[[(1,3)]]
        elif isinstance(self._indices[0][0],numbers.Number): #[(0,2),(1,2)]
            self._indices = [self._indices,] #->[[(0,2),(1,3)],]
        elif not isinstance(self._indices[0][0][0],numbers.Number): #[[(0,2),(1,2)]]
            raise TequilaException(f'Indices formating not recognized, received {self._indices}')
        if isinstance(self._variables,FixedVariable):
            assert 1 == len(self._indices) 
        else:
            assert len(self._variables) == len(self._indices)

    @property
    def qubits(self):
        if self._indices is not None:
            q = []
            for gate in self._indices:
                for exct in gate:
                    q.extend(exct)
            return sorted(list(set(q)))
        else: return 0

    @property
    def n_qubits(self)->int:
        return len(self.qubits)
    
    def map_qubits(self,qubit_map:dict={}):
        self._indices =[[(qubit_map[idx[0]],qubit_map[idx[1]]) for idx in gate] for gate in self._indices]
        return self

    def map_variables(self,var_map:dict={}):
        if self.is_parameterized() and hasattr(self.variables, "extract_variables"):
            self.variables = self.variables.map_variables(var_map)
        return self 

    def dagger(self):
        indinces = []
        cir = deepcopy(self)
        for gate in reversed(cir._indices):
            indinces.extend([[tuple([*idx][::-1]) for idx in gate]])
        cir.indices = indinces
        cir.variables = -1*self.variables
        return cir

    @property
    def max_qubit(self):
        if self.qubits:
            return self.qubits[-1]
        else: return 0

    def __eq__(self, other) -> bool:
        if self.name != other.name:
            return False
        if self.reordered != other.reordered:
            return False
        if self.variables != other.variables:
            return False
        if self.indices != other.indices:
            return False
        return True

class FermionicExcitationImpl(FGateImpl):
    def __init__(self, indices:typing.Union[list,tuple], variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable,None]=None, reordered:bool=False):
        super().__init__(indices, variables, reordered)
        self._name = 'FermionicExcitation'

class URImpl(FGateImpl):
    def __init__(self, i:int,j:int, variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable,None]=None):
        super().__init__([[(2*i,2*j)],[(2*i+1,2*j+1)]], variables, False)
        self._name = 'UR'
    def verify(self):
        assert isinstance(self._variables,(typing.Hashable, numbers.Real, Variable, FixedVariable))
        if isinstance(self._indices[0],numbers.Number): #[1,3]
            self._indices = [[tuple(self._indices),],] #->[[(1,3)]]
        elif isinstance(self._indices[0][0],numbers.Number): #[(0,2),(1,2)]
            self._indices = [self._indices,] #->[[(0,2),(1,3)],]
        elif not isinstance(self._indices[0][0][0],numbers.Number): #[[(0,2),(1,2)]]
            raise TequilaException(f'Indices formating not recognized, received {self._indices}')
        if isinstance(self._variables,FixedVariable):
            assert 2 == len(self._indices) 
        else:
            assert 2*len(self._variables) == len(self._indices)

class UCImpl(FGateImpl):
    def __init__(self,i,j, variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]=None):
        super().__init__([[(2*i,2*j),(2*i+1,2*j+1)]], variables, False)
        self._name = 'UC'
    
class PhaseImpl(FGateImpl):
    def __init__(self,i, variables:typing.Union[typing.Hashable, numbers.Real, Variable, FixedVariable]=None, reordered:bool=False):
        super().__init__([[(i,i)]], variables, reordered)
        self._name = 'Ph'

    def __str__(self):
        return f'{self.name}(target = {(self.indices[0][0][0],)}, variable = {repr(self.variables)})'



import copy
from typing import List, Optional

from ..core import Gate
import tequila as tq


class NamedGate(Gate):
    """
    Simple rectangular gate, with auto-sized width (if not explicitly passed)
    """

    def __init__(self, name: str, circuit: tq.QCircuit, wires: List[int], width: Optional[int] = None):
        self.name = name
        self.circuit = circuit
        self.width = width
        self.wires = wires

    def _render(self, state, style) -> str:
        gates = ""
        for gate in self.wires:
            gates += "a" + str(gate) + " "

        if self.width is None:
            self.width = 10 + len(self.name) * 3

        return gates + "G:width=" + str(self.width) + " {" + self.name + "}"

    def used_wires(self) -> List[int]:
        return self.wires

    def construct_circuit(self) -> tq.QCircuit:
        return self.circuit

    def map_variables(self, variables) -> "NamedGate":
        s = copy.deepcopy(self)
        s.circuit.map_variables(variables)
        return s

    def dagger(self) -> "NamedGate":
        return NamedGate(self.name + "^{\\dagger}", self.circuit.dagger(), self.wires, self.width)



from .tequila_expval import TequilaBraket
from typing import Union,List
from openfermion import FermionOperator
from tequila import TequilaException

SUPPORTED_FERMIONIC_BACKENDS = ["tequila", "fqe", "tcc"]
INSTALLED_FERMIONIC_BACKENDS = {"tequila": TequilaBraket}

try:
    from sunrise.expval.tcc_expval import TCCBraket
    INSTALLED_FERMIONIC_BACKENDS["tcc"] = TCCBraket
except ImportError:
    pass
try:
    from sunrise.expval.fqe_expval import FQEBraKet
    INSTALLED_FERMIONIC_BACKENDS["fqe"] = FQEBraKet
except ImportError:
    pass

def show_available_modules():
    print("Available Fermionic Modules:")
    for k in INSTALLED_FERMIONIC_BACKENDS.keys():
        print(k)


def show_supported_modules():
    print(SUPPORTED_FERMIONIC_BACKENDS)

def Braket(backend:str='tequila',operator:Union[str,List[FermionOperator],FermionOperator]=None,*args,**kwargs)->TequilaBraket:
    '''
    Interface for Fermionic Braket <U_bra|H|U_ket>. Since it allows U1 != U2, and U1=U2, some keywords can be presented as
    keywordX -> X \in [_bra,_ket,None] (i.e. init_state,init_state_ket,init_state_bra). If X=None, bra and ket are setted. 
    If X=ket, but not X=bra provided, bra is setted to ket.
    Parameters
    ----------
    backend
        fermionic backend to compute the Braket. See sun.show_supported_modules() or show_available_modules().
    mol/molecule
        tequila or pyscf molecule to get the molecular info and the electronic integrals
    integral_mager + parameters:
        alternative to molecule, tequila objects providing required molecular information
    h/int1e/one_body_integrals + g/int2e/two_body_integrals + (Optional)c/e_core/constant_term
    + (Optional) s/ovlp/overlap_integrals + n_elec
        electronic integral information
    (Optional)n_qubits
        number of circuit's qubits. If not provided, assumed from molecular information
    bra/ket/circuit: FCircuit
        Fermionic Circuit containing the excitations, variables and initial state.
    (Optional) backend_kwargs:dict
        tcc
            engine
                see tcc -> UCC engine
            dtype
                see tcc.set_dtype
            backend
                see tcc.set_backend
            else: kwargs provided to the UCC object initialization
        tequila: everything provided in tq.ExpVal/Braket furthen than the circuit and operators
    '''
    if 'H' in kwargs:
            H = kwargs['H']
            kwargs.pop('H')
            if operator is not None:
                raise TequilaException('Two operators provided?')
            else:
                operator = H
    if isinstance(operator,list):
        return [INSTALLED_FERMIONIC_BACKENDS[backend.lower()](operator=op,*args,**kwargs) for op in operator]
    #any kwargs and circuit form should be managed inside each class
    return INSTALLED_FERMIONIC_BACKENDS[backend.lower()](operator=operator,*args,**kwargs) 



from sunrise import fermionic_operations as fe
from tequila import QTensor,Variable,Objective,TequilaException,simulate
import numpy
import numbers
from typing import Union
# Adaptation of get_gives_decompostion from tequila qc_base at version 1.9.9dev at August 2025

OPTIMIZED_ORDERING = "Optimized"

def n_rotation(i:int, phi)->fe.FCircuit:
        """
        Creates a quantum circuit that applies a phase rotation based on phi to both components (up and down) of a given spatial orbital.

        Parameters:
        - i (int): The index of the qubit to which the rotation will be applied.
        - phi (float): The rotation angle. The actual rotation applied will be multiplied with -2 for both components.

        Returns:
        - FCircuit: A quantum circuit object containing the sequence of rotations applied to the up and down components of the specified spatial orbital.
        """

        # Start a new circuit and apply rotations to each component.
        circuit = fe.Phase(2*i, variables=-2 * phi)
        circuit += fe.Phase(2*i+1, variables=-2 * phi)
        return circuit

def get_givens_circuit(unitary:numpy.ndarray, tol:float=1e-6, ordering:Union[list,tuple,str]=OPTIMIZED_ORDERING)->fe.FCircuit:
    """
    Constructs a quantum circuit from a given real unitary matrix using Givens rotations.

    This method decomposes a unitary matrix into a series of Givens and Phase rotations,
    then constructs and returns a quantum circuit that implements this sequence of rotations.

    Parameters:
    - unitary (numpy.array): A real unitary matrix representing the transformation to implement.
    - tol (float): A tolerance threshold below which matrix elements are considered zero.
    - ordering (list of tuples or 'Optimized'): Custom ordering of indices for Givens rotations or 'Optimized' to generate them automatically.

    Returns:
    - FCircuit: A quantum circuit implementing the series of rotations decomposed from the unitary.
    """
    # Decompose the unitary matrix into Givens and phase (Rz) rotations.
    theta_list, phi_list = get_givens_decomposition(unitary, tol, ordering)

    # Initialize an empty quantum circuit.
    circuit = fe.FCircuit()

    # Add all Rz (phase) rotations to the circuit.
    for phi in phi_list:
        if  isinstance(phi[0],numbers.Number) and abs(phi[0])%(2*numpy.pi)<tol:
                continue
        circuit += n_rotation(phi[1], phi[0])

    # Add all Givens rotations to the circuit.
    for theta in reversed(theta_list):
        if  isinstance(theta[0],numbers.Number) and abs(theta[0])%numpy.pi<tol:
                continue
        circuit += fe.UR(theta[2], theta[1], theta[0] * -2)
    return circuit

def givens_matrix(n, p, q, theta)->QTensor:
    """
    Construct a complex Givens rotation matrix of dimension n by theta between rows/columns p and q.
    """
    """
    Generates a Givens rotation matrix of size n x n to rotate by angle theta in the (p, q) plane. This matrix can be complex

    Parameters:
    - n (int): The size of the Givens rotation matrix.
    - p (int): The first index for the rotation plane.
    - q (int): The second index for the rotation plane.
    - theta (float): The rotation angle.

    Returns:
    - numpy.array: The Givens rotation matrix.
    """
    matrix = QTensor(shape=(n, n), objective_list=numpy.eye(n).reshape(n * n))  # Matrix to hold complex numbers
    
    if isinstance(theta, (Variable, Objective)):
        cos_theta = theta.apply(numpy.cos)
        sin_theta = theta.apply(numpy.sin)
    else:
        cos_theta = numpy.cos(theta)
        sin_theta = numpy.sin(theta)

    # Directly assign cosine and sine without complex phase adjustment
    matrix[p, p] = cos_theta
    matrix[q, q] = cos_theta
    matrix[p, q] = sin_theta
    matrix[q, p] = -sin_theta

    return matrix

def get_givens_decomposition(unitary:numpy.ndarray, tol:float=1e-6, ordering:Union[list,tuple,str]=OPTIMIZED_ORDERING, return_diagonal:bool=False):
    """
    Decomposes a real unitary matrix into Givens rotations (theta) and Rz rotations (phi).

    Parameters:
    - unitary (numpy.array): A real unitary matrix to decompose. It cannot be complex.
    - tol (float): Tolerance for considering matrix elements as zero. Elements with absolute value less than tol are treated as zero.
    - ordering (list of tuples or 'Optimized'): Custom ordering of indices for Givens rotations or 'Optimized' to generate them automatically.
    - return_diagonal (bool): If True, the function also returns the diagonal matrix as part of the output.

    Returns:
    - list: A list of tuples, each representing a Givens rotation. Each tuple contains the rotation angle theta and indices (i,j) of the rotation.
    - list: A list of tuples, each representing an Rz rotation. Each tuple contains the rotation angle phi and the index (i) of the rotation.
    - numpy.array (optional): The diagonal matrix after applying all Givens rotations, returned if return_diagonal is True.
    """
    U = unitary  # no need to copy as we don't modify the original
    # U[abs(U) < tol] = 0 # Zeroing out the small elements as per the tolerance level. #comented out, its being considered latter again
    n = U.shape[0]

    # Determine optimized ordering if specified.
    if ordering == OPTIMIZED_ORDERING:
        ordering = depth_eff_order_mf(n)

    theta_list = []
    phi_list = []

    def calcTheta(U, c, r):
        """Calculate and apply the Givens rotation for a specific matrix element."""
        t = arctan2(-U[r, c], U[r - 1, c])
        if isinstance(t,numbers.Number) and numpy.isclose(numpy.abs(t)%numpy.pi,0,atol=tol):
            return U
        theta_list.append((t, r, r - 1))
        g = givens_matrix(n, r, r - 1, t)  # is a QTensor
        U = g.dot(U)
        return U

    # Apply and store Givens rotations as per the given or computed ordering.
    if ordering is None:
        for c in range(n):
            for r in range(n - 1, c, -1):
                U = calcTheta(U, c, r)
    else:
        for r, c in ordering:
            U = calcTheta(U, c, r)
    # Calculating the Rz rotations based on the phases of the diagonal elements.
    # For real elements this means a 180 degree shift, i.e. a sign change.
    for i in range(n):
        if isinstance(U[i, i], (Variable, Objective)):
            if len(U[i, i].args):
                phi_list.append((U[i, i].apply(numpy.angle), i))
        else:
            phi_list.append((numpy.angle(U[i, i]), i))
    # Filtering out rotations without significance.
    theta_list_new = []
    for i, theta in enumerate(theta_list):
        if isinstance(theta[0], (Variable, Objective)):
            if len(theta[0].args):
                theta_list_new.append(theta)
        elif (abs(theta[0])%( 2 * numpy.pi)) > tol:
            theta_list_new.append(theta)
    phi_list_new = []
    for i, phi in enumerate(phi_list):
        if isinstance(phi[0], (Variable, Objective)):
            if len(phi[0].args):
                phi_list_new.append(phi)
        elif abs(phi[0]) > tol:
            phi_list_new.append(phi)
    if return_diagonal:
        # Optionally return the resulting diagonal
        return theta_list_new, phi_list_new, U
    else:
        return theta_list_new, phi_list_new

def reconstruct_matrix_from_givens(n:int, theta_list:Union[list,tuple], phi_list:Union[list,tuple], to_real_if_possible:bool=True, tol:float=1e-12)->numpy.ndarray:
    """
    Reconstructs a matrix from given Givens rotations and Phase diagonal rotations.
    This function is effectively an inverse of get_givens_decomposition, and therefore only works with data in the same format as its output.

    Parameters:
    - n (int): The size of the unitary matrix to be reconstructed.
    - theta_list (list of tuples): Each tuple contains (angle, i, j) representing a Givens rotation of `angle` radians, applied to rows/columns `i` and `j`.
    - phi_list (list of tuples): Each tuple contains (angle, i), representing an Rz rotation by `angle` radians applied to the `i`th diagonal element.
    - to_real_if_possible (bool): If True, converts the matrix to real if its imaginary part is effectively zero.
    - tol (float): The tolerance whether to swap a complex rotation for a sign change.

    Returns:
    - numpy.ndarray: The reconstructed complex or real matrix, depending on the `to_real_if_possible` flag and matrix composition.
    """
    # Start with an identity matrix
    # reconstructed = numpy.eye(norb, dtype=complex)
    reconstructed =QTensor(shape=(n, n), objective_list=numpy.eye(n, dtype=complex).reshape(n * n))
    
    # Apply Rz rotations for diagonal elements
    for phi in phi_list:
        angle, i = phi
        if isinstance(angle, (Variable, Objective)):
            reconstructed[i, i] = reconstructed[i, i]*angle.apply(numpy.cos) + 1j*reconstructed[i, i]*angle.apply(numpy.sin)
        else:
            # Directly apply a sign flip if the rotation angle is π
            if numpy.isclose(angle, numpy.pi, atol=tol):
                reconstructed[i, i] *= -1
            else:
                reconstructed[i, i] *= numpy.exp(1j * angle)
    
    
    # Apply Givens rotations in reverse order
    for theta in reversed(theta_list):
        angle, i, j = theta
        g = givens_matrix(n, i, j, angle)
        reconstructed = reconstructed.dot(g)  

    # Convert matrix to real if its imaginary part is negligible unless disabled via to_real_if_possible
    if to_real_if_possible:
        # Directly apply a sign flip if the rotation angle is π
        if numpy.all(reconstructed.imag == 0):
            # Convert to real by taking the real part
            reconstructed = reconstructed.real

    return reconstructed.T

def reconstruct_matrix_from_circuit(U:fe.FCircuit,norb:int=None, to_real_if_possible:bool=True, tol:float=1e-12)->QTensor:#numpy.ndarray:
    """
    Reconstructs a matrix from a UR + Phase FCircuit.
    This function is effectively an inverse of get_givens_circuit.

    Parameters:
    - norb (int): The size of the unitary matrix to be reconstructed. If None, the number of qubits//2 in U is used.
    - U (FCircuit): The quantum circuit containing only UR (optionally Excitations gates) and Phase rotations.
    - to_real_if_possible (bool): If True, converts the matrix to real if its imaginary part is effectively zero.
    - tol (float): The tolerance whether to swap a complex rotation for a sign change.

    Returns:
    - numpy.ndarray: The reconstructed complex or real matrix, depending on the `to_real_if_possible` flag and matrix composition.
    """
    U = U.to_udud(norb)
    if norb is None:
        norb = U.n_qubits // 2
    phi_list = []
    theta_list = []  
    first = True
    for i,gate in enumerate(U.gates):
        if gate.name =='UR':
            theta_list.append((0.5*gate.variables, gate.indices[0][0][0]//2, gate.indices[0][0][1]//2))
        elif gate.name == "Ph":
            if not first:
                first = True
                continue
            if i+1!=len(U.gates) and U.gates[i+1].name == "Ph"  and U.gates[i+1].indices[0][0][0]//2 == gate.indices[0][0][0]//2:
                if isinstance(gate.variables, (Variable, Objective)) and isinstance(U.gates[i+1].variables, (Variable, Objective)): # Not sure if I can check the to Objectives to be the same, so I just go on
                    phi_list.append((-0.5*gate.variables, gate.indices[0][0][0]//2))
                elif isinstance(gate.variables, numbers.Number) and isinstance(U.gates[i+1].variables, numbers.Number) and numpy.isclose(gate.variables,U.gates[i+1].variables, atol=tol):
                    phi_list.append((-0.5*gate.variables, gate.indices[0][0][0]//2))
                else:
                    raise TequilaException("Phase gates must come in pairs to be converted to matrix.")
                first = False
            else:
                raise TequilaException("Phase gates must come in pairs to be converted to matrix.")
        elif gate.name == "FermionicExcitation":
            if not first:
                first = True
                continue
            if i+1!=len(U.gates) and U.gates[i+1].name == "FermionicExcitation" and len(U.gates[i+1].indices[0])==len(gate.indices[0])==1 and gate.indices[0][0][0]//2 == U.gates[i+1].indices[0][0][0]//2 and gate.indices[0][0][1]//2 == U.gates[i+1].indices[0][0][1]//2:
                if isinstance(gate.variables, (Variable, Objective)) and isinstance(U.gates[i+1].variables, (Variable, Objective)): # Not sure if I can check the to Objectives to be the same, so I just go on
                    theta_list.append((0.5*gate.variables, gate.indices[0][0][0]//2, gate.indices[0][0][1]//2))
                elif isinstance(gate.variables, numbers.Number) and isinstance(U.gates[i+1].variables, numbers.Number) and numpy.isclose(gate.variables, U.gates[i+1].variables, atol=tol):
                    theta_list.append((0.5*gate.variables, gate.indices[0][0][0]//2, gate.indices[0][0][1]//2))
                else:
                    raise TequilaException("FermionicExcitation gates must come in pairs to be converted to UR gates.")
                first = False
            else:
                raise TequilaException("FermionicExcitation gates must come in pairs to be converted to UR gates.")
        else:
            raise TequilaException(f"Reconstruction from circuits gate {gate} not implented.")
    return reconstruct_matrix_from_givens(norb, theta_list[::-1], phi_list[::-1], to_real_if_possible, tol)



def arctan2(x1, x2, *args, **kwargs):
    if isinstance(x1, (Variable, Objective)) or isinstance(x2, (Variable, Objective)):
        return Objective().binary_operator(left=1 * x1, right=1 * x2, op=numpy.arctan2)
    elif not isinstance(x1, numbers.Complex) and not isinstance(x2, numbers.Complex):
        return numpy.arctan2(x1, x2)
    else:
        return numpy.arctan2(x1.imag, x2.imag) + numpy.arctan2(x1.real, x2.real)


#This one copied from tequila.grouping.fermionic_functions, same date and patch as above
def depth_eff_order_mf(N:int)->list:
    """
    Returns index ordering for linear depth circuit

    For example N = 6 gives elimination order
    [ 0.  0.  0.  0.  0.  0.]
    [ 7.  0.  0.  0.  0.  0.]
    [ 5. 10.  0.  0.  0.  0.]
    [ 3.  8. 12.  0.  0.  0.]
    [ 2.  6. 11. 14.  0.  0.]
    [ 1.  4.  9. 13. 15.  0.]
    """
    l = []
    for c in range(0, N - 1):
        for r in range(1, N):
            if r - c > 0:
                l.append([r, c, 2 * c - r + N])
    l.sort(key=lambda x: x[2])
    return [(a[0], a[1]) for a in l]